from pyticketswitch.ticket_type import TicketType
from pyticketswitch.price_band import PriceBand
from pyticketswitch.discount import Discount
from pyticketswitch.seat import SeatBlock, Seat


class TestTicketType:

    def test_from_api_data(self):

        data = {
            'ticket_type_desc': 'Grand Circle',
            'price_band': [
                {
                    'price_band_desc': '',
                    'cost_range': {
                        'quantity_options': {
                            'valid_quantity_mask': 2046
                        },
                        'max_surcharge': 11.65,
                        'max_seatprice': 59.5,
                        'range_currency': {
                            'currency_factor': 100,
                            'currency_places': 2,
                            'currency_post_symbol': '',
                            'currency_number': 826,
                            'currency_pre_symbol': '\xa3',
                            'currency_code': 'gbp'
                        },
                        'min_surcharge': 11.25,
                        'no_singles_cost_range': {
                            'quantity_options': {
                                'valid_quantity_mask': 2046
                            },
                            'max_surcharge': 11.65,
                            'max_seatprice': 59.5,
                            'range_currency': {
                                'currency_factor': 100,
                                'currency_places': 2,
                                'currency_post_symbol': '',
                                'currency_number': 826,
                                'currency_pre_symbol': '\xa3',
                                'currency_code': 'gbp'
                            },
                            'min_surcharge': 11.25,
                            'min_seatprice': 57.5
                        },
                        'min_seatprice': 57.5
                    },
                    'price_band_code': 'B'
                },
            ],
            'ticket_type_code': 'GRA'
        }
        ticket_type = TicketType.from_api_data(data)
        assert ticket_type.code == 'GRA'
        assert ticket_type.description == 'Grand Circle'
        assert len(ticket_type.price_bands) == 1

    def test_get_seats(self):

        ticket_type = TicketType(
            'STALLS',
            price_bands=[
                PriceBand(
                    'A/pool',
                    Discount('ADULT'),
                    allows_leaving_single_seats='always',
                    seat_blocks=[
                        SeatBlock(
                            2,
                            seats=[
                                Seat(id_='A1'),
                                Seat(id_='A2'),
                            ]
                        ),
                        SeatBlock(
                            3,
                            seats=[
                                Seat(id_='B1'),
                                Seat(id_='B2'),
                                Seat(id_='B3'),
                            ]
                        ),
                    ]
                ),
                PriceBand(
                    'B/pool',
                    Discount('ADULT'),
                    allows_leaving_single_seats='always',
                    seat_blocks=[
                        SeatBlock(
                            1,
                            seats=[
                                Seat(id_='C1'),
                            ]
                        ),
                        SeatBlock(
                            2,
                            seats=[
                                Seat(id_='D1'),
                                Seat(id_='D2'),
                            ]
                        ),
                    ]
                ),
            ]
        )

        seats = ticket_type.get_seats()

        assert seats[0].id == 'A1'
        assert seats[1].id == 'A2'
        assert seats[2].id == 'B1'
        assert seats[3].id == 'B2'
        assert seats[4].id == 'B3'
        assert seats[5].id == 'C1'
        assert seats[6].id == 'D1'
        assert seats[7].id == 'D2'

    def test_get_seats_with_no_price_bands(self):

        ticket_type = TicketType(
            'STALLS',
            price_bands=None
        )

        seats = ticket_type.get_seats()

        assert seats == []



from pyticketswitch.customer import Customer


class TestCustomer:

    def test_from_api_data(self):

        data = {
            "addr_line_one": "Metro Building",
            "addr_line_one_latin": "METRO BUILDING",
            "addr_line_two": "1 Butterwick",
            "addr_line_two_latin": "1 BUTTERWICK",
            "agent_ref": "foobar123",
            "country": "United Kingdom",
            "country_code": "uk",
            "country_latin": "UNITED KINGDOM",
            "county": "London",
            "county_latin": "LONDON",
            "dp_supplier": True,
            "dp_user": True,
            "dp_world": True,
            "email_addr": "tester@gmail.com",
            "first_name": "Test",
            "first_name_latin": "TEST",
            "home_phone": "0203 137 7420",
            "work_phone": "079888888888",
            "phone": "079000000000",
            "initials": "l.o.l",
            "initials_latin": "L.O.L",
            "last_name": "Tester",
            "last_name_latin": "TESTER",
            "postcode": "w6 8dl",
            "postcode_latin": "W6 8DL",
            "suffix": "esq",
            "suffix_latin": "ESQ",
            "title": "rt.hon",
            "title_latin": "RT.HON",
            "town": "Hammersmith",
            "town_latin": "HAMMERSMITH",
        }

        customer = Customer.from_api_data(data)

        assert customer.first_name == 'Test'
        assert customer.first_name_latin == 'TEST'

        assert customer.last_name == 'Tester'
        assert customer.last_name_latin == 'TESTER'

        assert customer.address_lines == [
            'Metro Building',
            '1 Butterwick',
        ]

        assert customer.address_lines_latin == [
            'METRO BUILDING',
            '1 BUTTERWICK',
        ]

        assert customer.country_code == 'uk'

        assert customer.title == 'rt.hon'
        assert customer.title_latin == 'RT.HON'

        assert customer.initials == 'l.o.l'
        assert customer.initials_latin == 'L.O.L'

        assert customer.suffix == 'esq'
        assert customer.suffix_latin == 'ESQ'

        assert customer.email == 'tester@gmail.com'

        assert customer.post_code == 'w6 8dl'
        assert customer.post_code_latin == 'W6 8DL'

        assert customer.town == 'Hammersmith'
        assert customer.town_latin == 'HAMMERSMITH'

        assert customer.county == 'London'
        assert customer.county_latin == 'LONDON'

        assert customer.country == 'United Kingdom'
        assert customer.country_latin == 'UNITED KINGDOM'

        assert customer.phone == '079000000000'
        assert customer.home_phone == '0203 137 7420'
        assert customer.work_phone == '079888888888'

        assert customer.agent_reference == 'foobar123'

        assert customer.supplier_can_use_data is True
        assert customer.user_can_use_data is True
        assert customer.world_can_use_data is True

    def test_as_api_parameters(self):
        customer = Customer(
            first_name='Test',
            first_name_latin='TEST',
            last_name='Tester',
            last_name_latin='TESTER',
            address_lines=['Metro Building', '1 Butterwick'],
            address_lines_latin=['METRO BUILDING', '1 BUTTERWICK'],
            country_code='uk',
            title='rt.hon',
            title_latin='RT.HON',
            initials='l.o.l',
            initials_latin='L.O.L',
            suffix='esq',
            suffix_latin='ESQ',
            email='tester@gmail.com',
            post_code='w6 8dl',
            post_code_latin='W6 8DL',
            town='Hammersmith',
            town_latin='HAMMERSMITH',
            county='London',
            county_latin='LONDON',
            country='United Kingdom',
            country_latin='UNITED KINGDOM',
            phone='079000000000',
            home_phone='0203 137 7420',
            work_phone='079888888888',
            agent_reference='foobar123',
            supplier_can_use_data=True,
            user_can_use_data=True,
            world_can_use_data=True,
        )

        assert customer.as_api_parameters() == {
            'first_name': 'Test',
            'last_name': 'Tester',
            'address_line_one': 'Metro Building',
            'address_line_two': '1 Butterwick',
            'country_code': 'uk',
            'title': 'rt.hon',
            'initials': 'l.o.l',
            'suffix': 'esq',
            'email_address': 'tester@gmail.com',
            'postcode': 'w6 8dl',
            'town': 'Hammersmith',
            'county': 'London',
            'phone': '079000000000',
            'home_phone': '0203 137 7420',
            'work_phone': '079888888888',
            'agent_ref': 'foobar123',
            'supplier_can_use_customer_data': True,
            'user_can_use_customer_data': True,
            'world_can_use_customer_data': True,
        }



from decimal import Decimal
from pyticketswitch.event import Event
from pyticketswitch.performance import Performance
from pyticketswitch.order import TicketOrder, Order
from pyticketswitch.seat import Seat
from pyticketswitch.commission import Commission


class TestTicketOrder:

    def test_from_api_data(self):
        data = {
            'discount_code': 'ADULT',
            'discount_desc': 'Adult standard',
            'no_of_seats': 2,
            'sale_seatprice': 25,
            'sale_surcharge': 2.50,
            'total_sale_seatprice': 50,
            'total_sale_surcharge': 5,
            'seats': [
                {'full_id': 'ABC123'},
                {'full_id': 'DEF456'},
            ],
            "total_sale_combined": 55.0,
            'sale_combined_tax_component': 35.0,
            'sale_surcharge_tax_sub_component': 0.0,
            'total_sale_combined_tax_component': 35.0,
            'total_sale_surcharge_tax_sub_component': 0.0,
        }

        ticket_order = TicketOrder.from_api_data(data)

        assert ticket_order.code == 'ADULT'
        assert ticket_order.number_of_seats == 2
        assert ticket_order.description == 'Adult standard'
        assert isinstance(ticket_order.seatprice, (int, float))
        assert ticket_order.seatprice == 25.0
        assert isinstance(ticket_order.surcharge, (int, float))
        assert ticket_order.surcharge == 2.5
        assert isinstance(ticket_order.total_seatprice, (int, float))
        assert ticket_order.total_seatprice == 50.0
        assert isinstance(ticket_order.total_surcharge, (int, float))
        assert ticket_order.total_surcharge == 5.0
        assert len(ticket_order.seats) == 2
        assert ticket_order.seats[0].id == 'ABC123'
        assert ticket_order.seats[1].id == 'DEF456'

        assert isinstance(ticket_order.raw_combined_tax_component, (int, float))
        assert ticket_order.raw_combined_tax_component == 35.0
        assert isinstance(ticket_order.raw_surcharge_tax_sub_component, (int, float))
        assert ticket_order.raw_surcharge_tax_sub_component == 0.0
        assert isinstance(ticket_order.raw_total_combined_tax_component, (int, float))
        assert ticket_order.raw_total_combined_tax_component == 35.0
        assert isinstance(ticket_order.raw_total_surcharge_tax_sub_component, (int, float))
        assert ticket_order.raw_total_surcharge_tax_sub_component == 0.0
        assert isinstance(ticket_order.total_sale_combined, (int, float))
        assert ticket_order.total_sale_combined == 55.0

    def test_from_api_data_with_decimal(self):
        data = {
            'discount_code': 'ADULT',
            'discount_desc': 'Adult standard',
            'no_of_seats': 2,
            'sale_seatprice': 25,
            'sale_surcharge': Decimal('2.50'),
            'total_sale_seatprice': 50,
            'total_sale_surcharge': Decimal('5.00'),
            'seats': [
                {'full_id': 'ABC123'},
                {'full_id': 'DEF456'},
            ],
            'total_sale_combined': Decimal('55.0'),
            'sale_combined_tax_component': Decimal('35.0'),
            'sale_surcharge_tax_sub_component': Decimal('0.0'),
            'total_sale_combined_tax_component': Decimal('35.0'),
            'total_sale_surcharge_tax_sub_component': Decimal('0.0'),
        }

        ticket_order = TicketOrder.from_api_data(data)

        assert ticket_order.code == 'ADULT'
        assert ticket_order.number_of_seats == 2
        assert ticket_order.description == 'Adult standard'
        assert isinstance(ticket_order.seatprice, (int, Decimal))
        assert ticket_order.seatprice == Decimal('25.0')
        assert isinstance(ticket_order.surcharge, (int, Decimal))
        assert ticket_order.surcharge == Decimal('2.5')
        assert isinstance(ticket_order.total_seatprice, (int, Decimal))
        assert ticket_order.total_seatprice == Decimal('50.0')
        assert isinstance(ticket_order.total_surcharge, (int, Decimal))
        assert ticket_order.total_surcharge == Decimal('5.0')
        assert len(ticket_order.seats) == 2
        assert ticket_order.seats[0].id == 'ABC123'
        assert ticket_order.seats[1].id == 'DEF456'

        assert isinstance(ticket_order.raw_combined_tax_component, (int, Decimal))
        assert ticket_order.raw_combined_tax_component == Decimal('35.0')
        assert isinstance(ticket_order.raw_surcharge_tax_sub_component, (int, Decimal))
        assert ticket_order.raw_surcharge_tax_sub_component == Decimal('0.0')
        assert isinstance(ticket_order.raw_total_combined_tax_component, (int, Decimal))
        assert ticket_order.raw_total_combined_tax_component == Decimal('35.0')
        assert isinstance(ticket_order.raw_total_surcharge_tax_sub_component, (int, Decimal))
        assert ticket_order.raw_total_surcharge_tax_sub_component == Decimal('0.0')
        assert isinstance(ticket_order.total_sale_combined, (int, Decimal))
        assert ticket_order.total_sale_combined == Decimal('55.0')

    def test_repr(self):
        ticket_order = TicketOrder('abc123')
        assert repr(ticket_order) == '<TicketOrder abc123>'

    def test_combined_price(self):
        data = {
            'discount_code': 'ADULT',
            'discount_desc': 'Adult standard',
            'no_of_seats': 2,
            'sale_seatprice': 25,
            'sale_surcharge': 2.50,
            'total_sale_seatprice': 50,
            'total_sale_surcharge': 5,
            'seats': [
                {'full_id': 'ABC123'},
                {'full_id': 'DEF456'},
            ]
        }
        ticket_order = TicketOrder.from_api_data(data)
        assert ticket_order.combined_price() == 27.50

    def test_total_combined_price(self):
        data = {
            'discount_code': 'ADULT',
            'discount_desc': 'Adult standard',
            'no_of_seats': 2,
            'sale_seatprice': 25,
            'sale_surcharge': 2.50,
            'total_sale_seatprice': 50,
            'total_sale_surcharge': 5,
            'seats': [
                {'full_id': 'ABC123'},
                {'full_id': 'DEF456'},
            ]
        }
        ticket_order = TicketOrder.from_api_data(data)
        assert ticket_order.total_combined_price() == 55.0


class TestOrder:

    def test_from_api_data(self):

        data = {
            "backend_purchase_reference": 'GHI098',
            "backend_cancellation_reference": 'GHI098-cancelled',
            "external_management_url": 'www.example.com/some-more-decisions',
            "cancellation_status": "possible",
            "cancellation_comment": "",
            "event": {
                "event_id": "6IF",
            },
            "item_number": 1,
            "performance": {
                "perf_id": "6IF-A7N",
            },
            "price_band_code": "C/pool",
            "price_band_desc": "Band C",
            "seat_request_status": "got_all",
            "ticket_orders": {
                "ticket_order": [
                    {"discount_code": "ADULT"},
                    {"discount_code": "CHILD"},
                ]
            },
            "ticket_type_code": "CIRCLE",
            "ticket_type_desc": "Upper circle",
            "total_no_of_seats": 3,
            "total_sale_seatprice": 51,
            "total_sale_surcharge": 5.40,
            "requested_seat_ids": [
                'ABC123',
                'DEF456',
            ],
            'send_method': {
                'send_code': 'POST',
                'send_cost': 3.5,
                'send_desc': 'Post (UK & Ireland only)',
                'send_type': 'post',
                'permitted_countries': {
                    'country': [
                        {
                            'country_code': 'ie',
                            'country_desc': 'Ireland'
                        },
                        {
                            'country_code': 'uk',
                            'country_desc': 'United Kingdom'
                        }
                    ]
                }
            },
            'predicted_gross_commission': {
                'amount_including_vat': 4.0,
                'amount_excluding_vat': 4.5,
                'commission_currency_code': 'gbp',
            },
            'predicted_user_commission': {
                'amount_including_vat': 2.0,
                'amount_excluding_vat': 2.1,
                'commission_currency_code': 'gbp',
            },
            "total_sale_combined_tax_component": 35.0,
            "total_sale_surcharge_tax_sub_component": 0.0,
            "total_sale_combined": 56.40
        }

        order = Order.from_api_data(data)

        assert order.item == 1
        assert order.price_band_code == 'C/pool'
        assert order.price_band_description == 'Band C'
        assert order.ticket_type_code == 'CIRCLE'
        assert order.ticket_type_description == 'Upper circle'
        assert order.number_of_seats == 3
        assert order.total_seatprice == 51
        assert order.total_surcharge == 5.40
        assert order.seat_request_status == "got_all"
        assert order.backend_purchase_reference == 'GHI098'
        assert order.backend_cancellation_reference == 'GHI098-cancelled'
        assert order.cancellation_status == "possible"
        assert order.cancellation_comment == ""
        assert order.external_management_url == 'www.example.com/some-more-decisions'

        assert isinstance(order.event, Event)
        assert order.event.id == '6IF'
        assert isinstance(order.performance, Performance)
        assert order.performance.id == '6IF-A7N'

        assert len(order.ticket_orders) == 2
        assert order.ticket_orders[0].code == 'ADULT'
        assert order.ticket_orders[1].code == 'CHILD'

        assert len(order.requested_seat_ids) == 2
        assert order.requested_seat_ids[0] == 'ABC123'
        assert order.requested_seat_ids[1] == 'DEF456'

        assert order.total_including_send_cost() == (51 + 5.40 + 3.5)

        assert isinstance(order.gross_commission, Commission)
        assert order.gross_commission.including_vat == 4.0
        assert order.gross_commission.excluding_vat == 4.5
        assert order.gross_commission.currency_code == 'gbp'
        assert isinstance(order.user_commission, Commission)
        assert order.user_commission.including_vat == 2.0
        assert order.user_commission.excluding_vat == 2.1
        assert order.user_commission.currency_code == 'gbp'

        assert order.total_sale_combined_tax_component == 35.0
        assert order.total_sale_surcharge_tax_sub_component == 0.0
        assert order.total_sale_combined == 56.40

    def test_from_api_data_with_decimal(self):

        data = {
            "backend_purchase_reference": 'GHI098',
            "event": {
                "event_id": "6IF",
            },
            "item_number": 1,
            "performance": {
                "perf_id": "6IF-A7N",
            },
            "price_band_code": "C/pool",
            "price_band_desc": "Band C",
            "seat_request_status": "got_all",
            "ticket_orders": {
                "ticket_order": [
                    {"discount_code": "ADULT"},
                    {"discount_code": "CHILD"},
                ]
            },
            "ticket_type_code": "CIRCLE",
            "ticket_type_desc": "Upper circle",
            "total_no_of_seats": 3,
            "total_sale_seatprice": 51,
            "total_sale_surcharge": Decimal('5.40'),
            "requested_seat_ids": [
                'ABC123',
                'DEF456',
            ],
            'send_method': {
                'send_code': 'POST',
                'send_cost': Decimal('3.5'),
                'send_desc': 'Post (UK & Ireland only)',
                'send_type': 'post',
                'permitted_countries': {
                    'country': [
                        {
                            'country_code': 'ie',
                            'country_desc': 'Ireland'
                        },
                        {
                            'country_code': 'uk',
                            'country_desc': 'United Kingdom'
                        }
                    ]
                }
            },
            'predicted_gross_commission': {
                'amount_including_vat': Decimal('4.0'),
                'amount_excluding_vat': Decimal('4.5'),
                'commission_currency_code': 'gbp',
            },
            'predicted_user_commission': {
                'amount_including_vat': Decimal('2.0'),
                'amount_excluding_vat': Decimal('2.1'),
                'commission_currency_code': 'gbp',
            },
            "total_sale_combined_tax_component": Decimal('35.0'),
            "total_sale_surcharge_tax_sub_component": Decimal('0.0'),
            "total_sale_combined": Decimal("56.40")
        }

        order = Order.from_api_data(data)

        assert order.item == 1
        assert order.price_band_code == 'C/pool'
        assert order.price_band_description == 'Band C'
        assert order.ticket_type_code == 'CIRCLE'
        assert order.ticket_type_description == 'Upper circle'
        assert order.number_of_seats == 3
        assert order.total_seatprice == 51
        assert order.total_surcharge == Decimal('5.40')

        assert len(order.ticket_orders) == 2

        assert len(order.requested_seat_ids) == 2

        assert order.total_including_send_cost() == (51 + Decimal('5.40') + Decimal('3.5'))

        assert isinstance(order.gross_commission, Commission)
        assert order.gross_commission.including_vat == Decimal('4.0')
        assert order.gross_commission.excluding_vat == Decimal('4.5')
        assert order.gross_commission.currency_code == 'gbp'
        assert isinstance(order.user_commission, Commission)
        assert order.user_commission.including_vat == Decimal('2.0')
        assert order.user_commission.excluding_vat == Decimal('2.1')
        assert order.user_commission.currency_code == 'gbp'

        assert order.total_sale_combined_tax_component == Decimal('35.0')
        assert order.total_sale_surcharge_tax_sub_component == Decimal('0.0')
        assert order.total_sale_combined == Decimal('56.40')

    def test_from_api_data_with_send_method(self):

        data = {
            'item_number': 1,
            "send_method": {
                "send_code": "COBO",
            },
        }

        order = Order.from_api_data(data)

        assert order.send_method.code == 'COBO'

    def test_get_seats(self):
        ticket_order_one = TicketOrder('a', seats=[
            Seat('A1'), Seat('A2'), Seat('A3'),
        ])

        ticket_order_two = TicketOrder('b', seats=[
            Seat('B1'), Seat('B2'), Seat('B3'),
        ])

        order = Order(1, ticket_orders=[ticket_order_one, ticket_order_two])

        seats = order.get_seats()
        assert [seat.id for seat in seats] == [
            'A1', 'A2', 'A3', 'B1', 'B2', 'B3',
        ]

    def test_get_seat_ids(self):
        ticket_order_one = TicketOrder('a', seats=[
            Seat('A1'), Seat('A2'), Seat('A3'),
        ])

        ticket_order_two = TicketOrder('b', seats=[
            Seat('B1'), Seat('B2'), Seat('B3'),
        ])

        order = Order(1, ticket_orders=[ticket_order_one, ticket_order_two])

        seat_ids = order.get_seat_ids()
        assert seat_ids == ['A1', 'A2', 'A3', 'B1', 'B2', 'B3']

    def test_get_seat_ids_where_ids_are_missing(self):
        ticket_order_one = TicketOrder('a', seats=[
            Seat('A1'), Seat('A2'), Seat(None),
        ])

        ticket_order_two = TicketOrder('b', seats=[
            Seat('B1'), Seat(None), Seat('B3'),
        ])

        order = Order(1, ticket_orders=[ticket_order_one, ticket_order_two])

        seat_ids = order.get_seat_ids()
        assert seat_ids == ['A1', 'A2', 'B1', 'B3']

    def test_unique_seat_text(self):
        ticket_order_one = TicketOrder('a', seats=[
            Seat('A1', seat_text='Hell bad'), Seat('A2'), Seat('A3'),
        ])

        ticket_order_two = TicketOrder('b', seats=[
            Seat('B1', seat_text='Hell good'), Seat('B2'), Seat('B3'),
        ])

        order = Order(1, ticket_orders=[ticket_order_one, ticket_order_two])

        seat_text = order.unique_seat_text()
        assert 'Hell bad' in seat_text
        assert 'Hell good' in seat_text

    def test_get_seats_with_no_ticket_orders(self):
        order = Order(1, ticket_orders=[])
        assert order.get_seats() == []

    def test_repr(self):
        order = Order(1, ticket_orders=[])
        assert repr(order) == '<Order 1>'

    def test_from_api_data_without_commission(self):
        data = {
            'item_number': 1,
            "send_method": {
                "send_code": "COBO",
            },
        }

        order = Order.from_api_data(data)

        assert order.gross_commission is None
        assert order.user_commission is None

    def test_from_api_data_with_actual_commission(self):
        data = {
            "gross_commission": {
                "amount_excluding_vat": 18.75,
                "amount_including_vat": 22.5,
                "commission_currency_code": "gbp"
            },
            "item_number": 1,
            "user_commission": {
                "amount_excluding_vat": 12.5,
                "amount_including_vat": 15,
                "commission_currency_code": "gbp"
            }
        }
        order = Order.from_api_data(data)

        assert type(order.user_commission) is Commission
        assert order.user_commission.including_vat == 15
        assert order.user_commission.excluding_vat == 12.5
        assert order.user_commission.currency_code == "gbp"

        assert type(order.gross_commission) is Commission
        assert order.gross_commission.including_vat == 22.5
        assert order.gross_commission.excluding_vat == 18.75
        assert order.gross_commission.currency_code == "gbp"




from pyticketswitch.user import User


class TestUser:

    def test_from_api_data(self):

        data = {
            "b2b_attributes": {
                "address": {
                    "addr_line_1": "",
                    "addr_line_2": "",
                    "can_be_edited": True,
                    "country_code": "uk",
                    "county": "",
                    "day_phone": "",
                    "email_addr": "",
                    "eve_phone": "",
                    "town": "",
                },
                "allow_anon_clients": True,
                "has_clients": False,
            },
            "direct_suppliers": [],
            "fixed_codes": {
                "area_codes": [],
                "backend_systems": [],
                "event_codes": [],
                "venue_codes": [],
            },
            "user_groups": [
                {"controllable": True, "group_desc": "Demo product", "group_number": 0}
            ],
            "user_info": {
                "always_force_offline_avail": False,
                "always_mark_as_test": True,
                "always_needs_agent_ref": False,
                "always_needs_email_addr": False,
                "backend_group": "demo_internal_credit_group",
                "can_buy": True,
                "can_login": True,
                "can_report_on_same_owner_code_users": False,
                "can_see_non_pool_perfs": False,
                "content_group": "demo_content_group",
                "contracting_office": "ingresso_uk",
                "custom_from_email_addr": "",
                "default_country_code": "uk",
                "default_lang_code": "",
                "desired_currency": "",
                "is_b2b": True,
                "max_reserve_minutes": 15,
                "owner_code": "",
                "real_name": "Demonstration User",
                "reports_show_cust_data": True,
                "statement_descriptor": "TICKETSWITCH-USER",
                "style": "fixed-tabs",
                "sub_style": "styled-aff-default",
                "sub_sub_style": "b2b-users",
                "sub_user": "pyticketswitch-test",
                "upfront_data_scheme": "",
                "user_id": "demo",
            },
        }

        user = User.from_api_data(data)

        assert user.id == "demo"
        assert user.name == "Demonstration User"
        assert user.country == "uk"
        assert user.sub_user == "pyticketswitch-test"
        assert user.is_b2b is True
        assert user.statement_descriptor == "TICKETSWITCH-USER"
        assert user.backend_group == "demo_internal_credit_group"
        assert user.content_group == "demo_content_group"



import pytest
import datetime
from dateutil.tz import tzoffset
from decimal import Decimal
from pyticketswitch.mixins import JSONMixin, PaginationMixin, SeatPricingMixin


class TestJSONMixin:

    ZULU = tzoffset('ZULU', 0)

    class Foo(JSONMixin, object):

        def __init__(self, bar):
            self.bar = bar

    def test_with_none(self):
        obj = self.Foo(None)
        result = obj.__jsondict__()
        assert result == {}

    def test_with_empty(self):
        obj = self.Foo([])
        result = obj.__jsondict__()
        assert result == {}

    def test_with_none_with_hide_none_false(self):
        obj = self.Foo(None)
        result = obj.__jsondict__(hide_none=False)
        assert result == {'bar': None}

    def test_with_empty_with_hide_none_false(self):
        obj = self.Foo([])
        result = obj.__jsondict__(hide_none=False)
        assert result == {}

    def test_with_none_with_hide_empty_false(self):
        obj = self.Foo(None)
        result = obj.__jsondict__(hide_empty=False)
        assert result == {}

    def test_with_empty_with_hide_empty_false(self):
        obj = self.Foo([])
        result = obj.__jsondict__(hide_empty=False)
        assert result == {'bar': []}

    def test_normal_object(self):
        obj = self.Foo('hello world!')
        result = obj.__jsondict__()
        assert result == {'bar': 'hello world!'}

    def test_datetime(self):
        date = datetime.datetime(2017, 1, 25, 12, 39, 40, tzinfo=self.ZULU)
        obj = self.Foo(date)
        result = obj.__jsondict__()
        assert result == {'bar': '2017-01-25T12:39:40+00:00'}

    def test_date(self):
        obj = self.Foo(datetime.date(2017, 1, 25))
        result = obj.__jsondict__()
        assert result == {'bar': '2017-01-25'}

    def test_sub_json(self):
        subobj = self.Foo('hello world!')
        obj = self.Foo(subobj)
        result = obj.__jsondict__()
        assert result == {'bar': {'bar': 'hello world!'}}

    def test_list_of_normals(self):
        obj = self.Foo(['hello', 'world!'])
        result = obj.__jsondict__()
        assert result == {'bar': ['hello', 'world!']}

    def test_dict_of_normals(self):
        obj = self.Foo({'first': 'hello', 'second': 'world!'})
        result = obj.__jsondict__()
        assert result == {'bar': {'first': 'hello', 'second': 'world!'}}

    def test_list_of_subobjs(self):
        obj = self.Foo([self.Foo('hello'), self.Foo('world!')])
        result = obj.__jsondict__()
        assert result == {'bar': [{'bar': 'hello'}, {'bar': 'world!'}]}

    def test_dict_of_subobjs(self):
        obj = self.Foo({
            'first': self.Foo('hello'),
            'second': self.Foo('world!')
        })
        result = obj.__jsondict__()
        assert result == {
            'bar': {
                'first': {'bar': 'hello'},
                'second': {'bar': 'world!'}
            }
        }

    def test_decimal_as_json(self):
        obj = self.Foo(Decimal('44.1234'))
        result = obj.as_json()
        assert result == '{"bar": 44.1234}'

    def test_as_json(self):
        obj = self.Foo('hello world!')
        result = obj.as_json()
        assert result == '{"bar": "hello world!"}'

    def test_as_dict_for_json(self):
        obj = self.Foo('hello world!')
        result = obj.as_dict_for_json()
        assert result == {'bar': 'hello world!'}


class TestPaginationMixin:

    def test_from_api_data(self):

        data = {
            'results': {
                "paging_status": {
                    "page_length": 50,
                    "page_number": 2,
                    "pages_remaining": 3,
                    "results_remaining": 150,
                    "total_unpaged_results": 250,
                }
            }
        }

        class FakeBaseMeta(object):

            @classmethod
            def from_api_data(cls, data):
                return cls()

        class FakeMeta(PaginationMixin, FakeBaseMeta):
            pass

        meta = FakeMeta.from_api_data(data)

        assert meta.page_length == 50
        assert meta.page_number == 2
        assert meta.pages_remaining == 3
        assert meta.results_remaining == 150
        assert meta.total_results == 250

    def test_is_paginated_pages_remaining(self):
        meta = PaginationMixin(
            page_length=50,
            page_number=1,
            pages_remaining=15,
            results_remaining=700,
            total_results=750,
        )

        assert meta.is_paginated() is True

    def test_is_paginated_with_less_results_than_page(self):
        meta = PaginationMixin(
            page_length=50,
            page_number=1,
            pages_remaining=0,
            results_remaining=0,
            total_results=30,
        )

        assert meta.is_paginated() is False

    def test_is_paginated_on_last_page(self):
        meta = PaginationMixin(
            page_length=50,
            page_number=15,
            pages_remaining=0,
            results_remaining=0,
            total_results=750,
        )

        assert meta.is_paginated() is True

    def test_is_paginated_when_something_is_fucked(self):
        meta = PaginationMixin(
            page_length=None,
            page_number=None,
            pages_remaining=None,
            results_remaining=None,
            total_results=None,
        )

        assert meta.is_paginated() is False

    def test_from_api_data_when_not_inside_results(self):
        """Test from_api_data works when the data is not in `results`

        Some calls (e.g. related_events.v1) have multiple pagination data
        sections inside dicts that are not keyed as 'results'. In this case,
        we can pass the subdict they are in directly, so we should make sure
        the method can find the data in the base dictionary.
        """

        # state

        data = {
            'misnamed_results': {
                'paging_status': {
                    'page_length': 50,
                    'page_number': 2,
                    'pages_remaining': 3,
                    'results_remaining': 150,
                    'total_unpaged_results': 250,
                }
            }
        }

        class FakeBaseMeta(object):

            @classmethod
            def from_api_data(cls, data):
                return cls()

        class FakeMeta(PaginationMixin, FakeBaseMeta):
            pass

        meta = FakeMeta.from_api_data(data, result_key='misnamed_results')

        assert meta.page_length == 50
        assert meta.page_number == 2
        assert meta.pages_remaining == 3
        assert meta.results_remaining == 150
        assert meta.total_results == 250


class TestSeatPricingMixin:

    def test_kwargs_from_api_data(self):

        data = {
            'sale_seatprice': 160,
            'sale_surcharge': 5.5,
            'non_offer_sale_seatprice': 200,
            'non_offer_sale_surcharge': 5.5,
        }

        kwargs = SeatPricingMixin.kwargs_from_api_data(data)

        assert kwargs['seatprice'] == 160.00
        assert kwargs['surcharge'] == 5.5
        assert kwargs['non_offer_seatprice'] == 200
        assert kwargs['non_offer_surcharge'] == 5.5

    def test_combined_price(self):
        inst = SeatPricingMixin(seatprice=123.45, surcharge=6.78)
        assert inst.combined_price() == 130.23

    def test_combined_price_missing_prices(self):
        inst = SeatPricingMixin(seatprice=123.45)
        with pytest.raises(AssertionError):
            inst.combined_price()

        inst = SeatPricingMixin(surcharge=6.78)
        with pytest.raises(AssertionError):
            inst.combined_price()

    def test_combined_price_inexact_floats(self):
        inst = SeatPricingMixin(seatprice=1.1, surcharge=2.2)
        assert inst.combined_price() == 3.3

    def test_combined_price_decimal(self):
        inst = SeatPricingMixin(
            seatprice=Decimal('123.45'),
            surcharge=Decimal('6.78')
        )
        assert inst.combined_price() == Decimal('130.23')

    def test_non_offer_combined_price(self):
        inst = SeatPricingMixin(non_offer_seatprice=123.45,
                                non_offer_surcharge=6.78)

        assert inst.non_offer_combined_price() == 130.23

    def test_non_offer_combined_price_missing_prices(self):
        inst = SeatPricingMixin(non_offer_seatprice=123.45)
        with pytest.raises(AssertionError):
            inst.non_offer_combined_price()

        inst = SeatPricingMixin(non_offer_surcharge=6.78)
        with pytest.raises(AssertionError):
            inst.non_offer_combined_price()

    def test_non_offer_combined_price_inexact_floats(self):
        inst = SeatPricingMixin(non_offer_seatprice=1.1, non_offer_surcharge=2.2)
        assert inst.non_offer_combined_price() == 3.3

    def test_non_offer_combined_price_decimal(self):
        inst = SeatPricingMixin(
            non_offer_seatprice=Decimal('123.45'),
            non_offer_surcharge=Decimal('6.78')
        )
        assert inst.non_offer_combined_price() == Decimal('130.23')



import json
import pytest

from pyticketswitch.cancellation import CancellationResult


class TestCancellationResult:
    def test_from_api_data_successful_cancellation(self):
        with open("test_data/successful_cancellation.json", 'r') as file_handle:
            data = json.load(file_handle)
        cancellation_result = CancellationResult.from_api_data(data)

        assert cancellation_result.cancelled_item_numbers == [1]
        assert len(cancellation_result.trolley.bundles) == 1
        assert len(cancellation_result.trolley.bundles[0].orders) == 1
        assert cancellation_result.must_also_cancel is None

    def test_from_api_data_must_also_cancel(self):
        with open("test_data/must_also_cancel_cancellation.json", 'r') as file_handle:
            data = json.load(file_handle)
        cancellation_result = CancellationResult.from_api_data(data)

        assert len(cancellation_result.cancelled_item_numbers) == 0
        assert len(cancellation_result.trolley.bundles) == 0
        assert len(cancellation_result.must_also_cancel) == 1
        assert cancellation_result.must_also_cancel[0].item == 2

    @pytest.mark.parametrize(
        "cancellation_data_file,expected_result",
        [
            ("test_data/successful_cancellation.json", True),
            ("test_data/must_also_cancel_cancellation.json", False),
            ("test_data/partial_cancellation.json", False),
            ("test_data/failed_cancellation.json", False),
        ],
    )
    def test_is_fully_cancelled(self, cancellation_data_file, expected_result):
        with open(cancellation_data_file, 'r') as file_handle:
            data = json.load(file_handle)
        cancellation_result = CancellationResult.from_api_data(data)

        assert cancellation_result.is_fully_cancelled() == expected_result



from pyticketswitch.price_band import PriceBand
from pyticketswitch.discount import Discount
from pyticketswitch.seat import SeatBlock, Seat


class TestPriceBand:

    def test_from_api_data(self):

        data = {
            'price_band_desc': 'Cheap Seats',
            'cost_range': {
                'quantity_options': {
                    'valid_quantity_mask': 2046
                },
                'max_surcharge': 11.65,
                'max_seatprice': 59.5,
                'range_currency': {
                    'currency_factor': 100,
                    'currency_places': 2,
                    'currency_post_symbol': '',
                    'currency_number': 826,
                    'currency_pre_symbol': '\xa3',
                    'currency_code': 'gbp'
                },
                'min_surcharge': 11.25,
                'no_singles_cost_range': {
                    'quantity_options': {
                        'valid_quantity_mask': 2046
                    },
                    'max_surcharge': 11.65,
                    'max_seatprice': 59.5,
                    'range_currency': {
                        'currency_factor': 100,
                        'currency_places': 2,
                        'currency_post_symbol': '',
                        'currency_number': 826,
                        'currency_pre_symbol': '\xa3',
                        'currency_code': 'gbp'
                    },
                    'min_surcharge': 11.25,
                    'min_seatprice': 57.5
                },
                'min_seatprice': 57.5
            },
            'price_band_code': 'B',
            'discount_code': 'ABC123',
            'allows_leaving_single_seats': 'always',
            'example_seats': [
                {'full_id': 'ZZ-TOP'},
                {'full_id': 'ZZ-BOTTOM'},
            ],
            'example_seats_are_real': False,
            'free_seat_blocks': {
                'blocks_by_row': {
                    'A': [['A1', 'A2'], ['A4', 'A5']],
                    'B': [['B-1', 'B-2'], ['B-4', 'B-5']],
                },
                'separators_by_row': {
                    'A': '',
                    'B': '-',
                },
                'seats_by_text_message': {
                    'Death Trap': ['A1', 'B-2'],
                },
                'restricted_view_seats': ['A1', 'B-2'],
            },
            "predicted_user_commission": {
                "amount_excluding_vat": 2.93,
            },
            "possible_discounts": {
                "discount": [
                    {
                        "discount_code": "ADULT",
                    },
                    {
                        "discount_code": "CHILD",
                    },
                ]
            },
            "number_available": 40,
            'sale_seatprice': 160,
            'sale_surcharge': 5.5,
            'non_offer_sale_seatprice': 200,
            'non_offer_sale_surcharge': 6.5,
            "percentage_saving": 20,
            "is_offer": True,
            'sale_combined_tax_component': 50
        }

        price_band = PriceBand.from_api_data(data)

        assert price_band.code == 'B'
        assert price_band.description == 'Cheap Seats'
        assert price_band.default_discount.code == 'ABC123'
        assert price_band.allows_leaving_single_seats == 'always'

        assert len(price_band.example_seats) == 2
        assert price_band.example_seats[0].id == 'ZZ-TOP'
        assert price_band.example_seats[1].id == 'ZZ-BOTTOM'
        assert price_band.example_seats_are_real is False

        assert len(price_band.seat_blocks) == 4
        assert price_band.seat_blocks[0].length == 2

        assert price_band.user_commission.excluding_vat == 2.93

        assert len(price_band.discounts) == 2
        assert price_band.discounts[0].code == 'ADULT'
        assert price_band.discounts[1].code == 'CHILD'

        assert price_band.availability == 40

        assert price_band.seatprice == 160.00
        assert price_band.surcharge == 5.5
        assert price_band.non_offer_seatprice == 200
        assert price_band.non_offer_surcharge == 6.5
        assert price_band.percentage_saving == 20
        assert price_band.is_offer is True
        assert price_band.tax_component == 50

    def test_get_seats(self):

        price_band = PriceBand(
            'A/pool',
            Discount('ADULT'),
            allows_leaving_single_seats='always',
            seat_blocks=[
                SeatBlock(
                    2,
                    seats=[
                        Seat(id_='A1'),
                        Seat(id_='A2'),
                    ]
                ),
                SeatBlock(
                    3,
                    seats=[
                        Seat(id_='B1'),
                        Seat(id_='B2'),
                        Seat(id_='B3'),
                    ]
                ),
            ]
        )

        seats = price_band.get_seats()

        assert seats[0].id == 'A1'
        assert seats[1].id == 'A2'
        assert seats[2].id == 'B1'
        assert seats[3].id == 'B2'
        assert seats[4].id == 'B3'

    def test_get_seats_with_seatblocks_without_seats(self):

        price_band = PriceBand(
            'A/pool',
            Discount('ADULT'),
            allows_leaving_single_seats='always',
            seat_blocks=[
                SeatBlock(
                    2,
                    seats=None
                ),
                SeatBlock(
                    3,
                    seats=[
                        Seat(id_='B1'),
                        Seat(id_='B2'),
                        Seat(id_='B3'),
                    ]
                ),
            ]
        )

        seats = price_band.get_seats()

        assert seats[0].id == 'B1'
        assert seats[1].id == 'B2'
        assert seats[2].id == 'B3'

    def test_get_seats_without_seatblocks(self):

        price_band = PriceBand(
            'A/pool',
            Discount('ADULT'),
            seat_blocks=None,
            allows_leaving_single_seats='always',
        )

        seats = price_band.get_seats()

        assert seats == []



from decimal import Decimal
from pyticketswitch.send_method import SendMethod


class TestSendMethod:

    def test_from_api_data(self):
        data = {
            'send_code': 'POST',
            'send_cost': 3.5,
            'send_desc': 'Post (UK & Ireland only)',
            'send_type': 'post',
            'send_final_comment': 'allow some time',
            'permitted_countries': {
                'country': [
                    {
                        'country_code': 'ie',
                        'country_desc': 'Ireland'
                    },
                    {
                        'country_code': 'uk',
                        'country_desc': 'United Kingdom'
                    }
                ]
            },
            'send_cost_tax_component': 0.95,
            'trans_fee_component': 1.5
        }

        method = SendMethod.from_api_data(data)

        assert isinstance(method, SendMethod)
        assert method.code == 'POST'
        assert method.cost == 3.5
        assert method.description == 'Post (UK & Ireland only)'
        assert method.type == 'post'
        assert method.final_comment == 'allow some time'
        assert len(method.permitted_countries) == 2
        assert method.send_cost_tax_component == 0.95
        assert method.trans_fee_component == 1.5

    def test_from_api_data_with_decimal(self):
        data = {
            'send_code': 'POST',
            'send_cost': Decimal('3.5'),
            'send_desc': 'Post (UK & Ireland only)',
            'send_type': 'post',
            'send_final_comment': 'allow some time',
            'permitted_countries': {
                'country': [
                    {
                        'country_code': 'ie',
                        'country_desc': 'Ireland'
                    },
                    {
                        'country_code': 'uk',
                        'country_desc': 'United Kingdom'
                    }
                ]
            },
            'send_cost_tax_component': Decimal('0.95'),
            'trans_fee_component': Decimal('1.5')
        }

        method = SendMethod.from_api_data(data)

        assert isinstance(method, SendMethod)
        assert method.code == 'POST'
        assert method.cost == Decimal('3.5')
        assert method.description == 'Post (UK & Ireland only)'
        assert method.type == 'post'
        assert method.final_comment == 'allow some time'
        assert len(method.permitted_countries) == 2
        assert method.send_cost_tax_component == Decimal('0.95')
        assert method.trans_fee_component == Decimal('1.5')

    def test_from_api_data_without_permitted_countries(self):
        data = {
            'send_code': 'COBO',
            'send_cost': 1.5,
            'send_desc': 'Collect from the venue',
            'send_type': 'collect',
        }

        method = SendMethod.from_api_data(data)

        assert isinstance(method, SendMethod)
        assert method.code == 'COBO'
        assert method.cost == 1.5
        assert method.description == 'Collect from the venue'
        assert method.type == 'collect'
        assert method.permitted_countries is None



import pytest
import datetime
import decimal
from datetime import date
from dateutil.tz import tzoffset
from decimal import Decimal
from pyticketswitch import utils
from pyticketswitch import exceptions


class TestDateRangeStr:

    def test_date_range_str(self):
        start_date = date(2016, 6, 21)
        end_date = date(2017, 1, 1)

        date_range_str = utils.date_range_str(start_date, end_date)
        assert date_range_str == '20160621:20170101'

    def test_date_range_str_with_datetimes(self):
        start_date = datetime.datetime(2016, 6, 21, 19, 30, 15)
        end_date = datetime.datetime(2017, 1, 1, 13, 45, 30)

        date_range_str = utils.date_range_str(start_date, end_date)
        assert date_range_str == '20160621:20170101'

    def test_date_range_str_with_no_end_date(self):
        start_date = date(2016, 6, 21)
        end_date = None

        date_range_str = utils.date_range_str(start_date, end_date)
        assert date_range_str == '20160621:'

    def test_date_range_str_with_no_start_date(self):
        start_date = None
        end_date = date(2017, 1, 1)

        date_range_str = utils.date_range_str(start_date, end_date)
        assert date_range_str == ':20170101'

    def test_date_range_str_with_no_start_date_or_end_date(self):
        start_date = None
        end_date = None

        date_range_str = utils.date_range_str(start_date, end_date)
        assert date_range_str == ''

    def test_date_range_str_with_invalid_end_date(self):
        start_date = date(2016, 6, 21)
        end_date = 'FOOBAR!'

        with pytest.raises(exceptions.InvalidParametersError):
            utils.date_range_str(start_date, end_date)

    def test_date_range_str_with_invalid_start_date(self):
        start_date = 'SAUSAGES!'
        end_date = date(2017, 1, 1)
        with pytest.raises(exceptions.InvalidParametersError):
            utils.date_range_str(start_date, end_date)


class TestIsoStrToDatetime:

    BST = tzoffset('BST', 3600)
    ZULU = tzoffset('ZULU', 0)

    def test_with_core_iso(self):
        date_str = '2016-09-16T19:30:00+01:00'
        dt = utils.isostr_to_datetime(date_str)

        assert dt == datetime.datetime(2016, 9, 16, 19, 30, 0, tzinfo=self.BST)

    def test_with_core_zulu(self):
        date_str = '2016-09-16T19:30:00Z'
        dt = utils.isostr_to_datetime(date_str)

        assert dt == datetime.datetime(2016, 9, 16, 19, 30, 0, tzinfo=self.ZULU)

    def test_with_python_iso(self):
        date_str = '2016-09-16T19:30:00+0100'
        dt = utils.isostr_to_datetime(date_str)

        assert dt == datetime.datetime(2016, 9, 16, 19, 30, 0, tzinfo=self.BST)

    def test_with_not_datetime(self):
        date_str = 'When the moon is in the forth corner of the jelly bean'
        with pytest.raises(ValueError):
            utils.isostr_to_datetime(date_str)

    def test_with_none(self):
        date_str = None
        with pytest.raises(ValueError):
            utils.isostr_to_datetime(date_str)

    def test_with_empty(self):
        date_str = ''
        with pytest.raises(ValueError):
            utils.isostr_to_datetime(date_str)


class TestYYYYToDate:

    def test_yyyymmdd_to_date_valid_string(self):
        date = utils.yyyymmdd_to_date('20160801')
        assert date.year == 2016
        assert date.month == 8
        assert date.day == 1

    def test_yyyymmdd_to_date_invalid_string(self):
        with pytest.raises(TypeError):
            utils.yyyymmdd_to_date(123)

        with pytest.raises(ValueError):
            utils.yyyymmdd_to_date('')

        with pytest.raises(ValueError):
            utils.yyyymmdd_to_date('wrong_date')


class TestSpecificDatesFromAPI:

    def test_specific_dates_from_api(self):
        api_data = {
            'year_2016': {
                'nov': {
                    'day_30': False,
                    'day_18': False,
                },
                'oct': {
                    'day_4': True,
                    'day_3': True,
                    'day_2': False,
                    'day_1': True,
                }
            }
        }
        results = utils.specific_dates_from_api_data(api_data)
        assert len(results) == 3
        assert type(results[0]) == datetime.date


class TestBitmaskToBooleanList:
    """
    NOTE: we are expecting big endian masks so the last bit of of our mask
    should the first element in our array
    """

    def test_simple_masks(self):
        # 0 == 0b0 so we only have one bit an it's a 0
        assert utils.bitmask_to_boolean_list(0) == [False]
        # 1 == 0b1 so we only have one bit an it's a 1
        assert utils.bitmask_to_boolean_list(1) == [True]
        # 5 == 0b101 so we expecting a length of 3 with the bits 1, 0 and 1
        assert utils.bitmask_to_boolean_list(5) == [True, False, True]
        # 6 == 0b110 so we expecting a length of 3 with the bits 0, 1 and 1
        assert utils.bitmask_to_boolean_list(6) == [False, True, True]


class TestBitmaskToNumberedList:

    def test_simple_masks(self):
        # 0 == 0b0 so we only have one bit an it's a 0
        assert utils.bitmask_to_numbered_list(0) == []
        # 1 == 0b1 so we only hanumberedbit an it's a 1
        assert utils.bitmask_to_numbered_list(1) == [1]
        # 5 == 0b101 so the 1st numbered bitst are 1's
        assert utils.bitmask_to_numbered_list(5) == [1, 3]
        # 6 == 0b110 so the 2nd numbered bits are 1's
        assert utils.bitmask_to_numbered_list(6) == [2, 3]

    def test_with_none(self):
        assert utils.bitmask_to_numbered_list(None) == []


class TestGetPrice:

    def test_with_keys(self):

        data = {
            'min_price': 12.40,
            'max_price': 22.00,
        }

        min_price = utils.get_price(data, 'min_price')
        assert min_price == 12.40

        max_price = utils.get_price(data, 'max_price')
        assert max_price == 22.00

    def test_with_float(self):

        data = {'price': 12.40}

        price = utils.get_price(data, 'price')
        assert isinstance(price, float)
        assert price == 12.40

    def test_with_ints(self):

        data = {'price': 12}

        price = utils.get_price(data, 'price')
        assert isinstance(price, float)
        assert price == 12.0

    def test_with_str_int(self):

        data = {'price': '12'}

        price = utils.get_price(data, 'price')
        assert isinstance(price, float)
        assert price == 12.0

    def test_with_str_float(self):

        data = {'price': '12.40'}

        price = utils.get_price(data, 'price')
        assert isinstance(price, float)
        assert price == 12.40

    def test_with_None(self):

        data = {'price': None}

        price = utils.get_price(data, 'price')
        assert price is None

    def test_with_missing_key(self):

        data = {'price': 12.40}

        max_price = utils.get_price(data, 'max_price')
        assert max_price is None

    def test_with_zero(self):

        data = {'price': 0}

        price = utils.get_price(data, 'price')
        assert isinstance(price, float)
        assert price == 0.0

    def test_with_str_zero(self):

        data = {'price': '0'}

        price = utils.get_price(data, 'price')
        assert isinstance(price, float)
        assert price == 0


class TestAddPrices:

    def test_adding_two_floats(self):
        combined = utils.add_prices(1.0, 1.0)
        assert combined == 2.0

    def test_adding_two_decimals(self):
        combined = utils.add_prices(Decimal('1.0'), Decimal('1.0'))
        assert combined == Decimal('2.0')

    def test_adding_two_ints(self):
        combined = utils.add_prices(1, 1)
        assert combined == 2

    def test_adding_two_strs(self):
        combined = utils.add_prices('1.0', '1.0')
        assert combined == '2.0'

    def test_adding_inexact_floats(self):
        combined = utils.add_prices(1.1, 2.2)
        assert combined == 3.3

    def test_adding_three_floats(self):
        combined = utils.add_prices(1.0, 1.0, 1.0)
        assert combined == 3.0

    def test_adding_a_decimal_and_a_float(self):
        combined = utils.add_prices(Decimal('1.0'), 1.0)
        assert combined == Decimal('2.0')

    def test_adding_a_float_and_a_decimal(self):
        combined = utils.add_prices(1.0, Decimal('1.0'))
        assert combined == Decimal('2.0')

    def test_adding_a_float_to_none(self):
        with pytest.raises(TypeError):
            utils.add_prices(1.0, None)

    def test_adding_a_decimal_to_none(self):
        with pytest.raises(TypeError):
            utils.add_prices(Decimal('1.0'), None)

    def test_zero_arguments(self):
        with pytest.raises(TypeError):
            utils.add_prices()

    def test_one_argument(self):
        with pytest.raises(TypeError):
            utils.add_prices(1)

    def test_invalid_string_price(self):
        with pytest.raises(decimal.InvalidOperation):
            utils.add_prices('not a price', 5)


class TestFilterNoneFromParameters:

    def test_with_parameters(self):

        params = {
            'foo': 'bar',
            'thing': None,
            'lol': 'beans',
        }

        assert utils.filter_none_from_parameters(params) == {
            'foo': 'bar',
            'lol': 'beans',
        }



from pyticketswitch.review import Review


class TestReview:

    def test_from_api_data(self):

        data = {
            'review_body': 'BODY',
            'review_iso8601_date_and_time': '2013-03-30T12:00:00Z',
            'star_rating': 4,
            'review_lang': 'en',
            'review_title': '',
            'is_user_review': False,
            'review_author': "What's on Stage",
            'review_original_url': 'http://www.whatsonstage.com/index.php?pg=207&story=E01822737494'
        }

        review = Review.from_api_data(data)
        assert review.body == 'BODY'
        assert review.date_time.date().year == 2013
        assert review.date_time.date().month == 3
        assert review.date_time.date().day == 30
        assert review.star_rating == 4
        assert review.language == 'en'
        assert review.author == "What's on Stage"
        assert review.url == 'http://www.whatsonstage.com/index.php?pg=207&story=E01822737494'
        assert review.is_user is False



from pyticketswitch.discount import Discount
from pyticketswitch.commission import Commission


class TestDiscount:

    def test_from_api_data(self):
        gross_commission_data = {
            'amount_including_vat': 2.8,
            'amount_excluding_vat': 3.0,
            'commission_currency_code': 'gbp',
        }
        user_commission_data = {
            'amount_including_vat': 0.8,
            'amount_excluding_vat': 1.0,
            'commission_currency_code': 'gbp',
        }

        data = {
            'discount_code': 'ADULT',
            'discount_desc': 'Adult standard',
            'discount_disallowed_seat_nos': [0, 2, 3, 4],
            'price_band_code': 'A/pool',
            'sale_seatprice': 160,
            'sale_surcharge': 5.5,
            'is_offer': True,
            'non_offer_sale_seatprice': 200,
            'non_offer_sale_surcharge': 6.5,
            'absolute_saving': 40,
            'percentage_saving': 20,
            'number_available': 6,
            'predicted_gross_commission': gross_commission_data,
            'predicted_user_commission': user_commission_data,
            'sale_combined_tax_component': 50,
            'discount_semantic_type': 'senior',
            'discount_minimum_eligible_age': 60,
            'discount_maximum_eligible_age': 80,
        }

        discount = Discount.from_api_data(data)

        assert discount.code == 'ADULT'
        assert discount.description == 'Adult standard'
        assert discount.disallowed_seat_nos == [0, 2, 3, 4]
        assert discount.price_band_code == 'A/pool'
        assert discount.seatprice == 160.0
        assert discount.surcharge == 5.5
        assert discount.is_offer is True
        assert discount.non_offer_seatprice == 200
        assert discount.non_offer_surcharge == 6.5
        assert discount.availability == 6
        assert isinstance(discount.gross_commission, Commission)
        assert discount.gross_commission.including_vat == gross_commission_data['amount_including_vat']
        assert discount.gross_commission.excluding_vat == gross_commission_data['amount_excluding_vat']
        assert discount.gross_commission.currency_code == gross_commission_data['commission_currency_code']
        assert isinstance(discount.user_commission, Commission)
        assert discount.user_commission.including_vat == user_commission_data['amount_including_vat']
        assert discount.user_commission.excluding_vat == user_commission_data['amount_excluding_vat']
        assert discount.user_commission.currency_code == user_commission_data['commission_currency_code']
        assert discount.tax_component == 50
        assert discount.semantic_type == 'senior'

    def test_from_api_data_without_commission(self):
        data = {
            'discount_code': 'ADULT',
            'discount_desc': 'Adult standard',
            'price_band_code': 'A/pool',
            'sale_seatprice': 160,
            'sale_surcharge': 5.5,
            'is_offer': True,
            'non_offer_sale_seatprice': 200,
            'non_offer_sale_surcharge': 6.5,
            'absolute_saving': 40,
            'percentage_saving': 20,
            'number_available': 6,
        }

        discount = Discount.from_api_data(data)

        assert discount.code == 'ADULT'
        assert discount.description == 'Adult standard'
        assert discount.price_band_code == 'A/pool'
        assert discount.seatprice == 160.0
        assert discount.surcharge == 5.5
        assert discount.is_offer is True
        assert discount.non_offer_seatprice == 200
        assert discount.non_offer_surcharge == 6.5
        assert discount.availability == 6
        assert discount.gross_commission is None
        assert discount.user_commission is None



from enum import Enum


MONTH_NUMBERS = {
    'jan': 1, 'feb': 2, 'mar': 3, 'apr': 4, 'may': 5, 'jun': 6, 'jul': 7,
    'aug': 8, 'sep': 9, 'oct': 10, 'nov': 11, 'dec': 12,
}


class DiscountSemanticType(Enum):
    BLUELIGHT = "bluelight"
    CARER = "carer"
    CHILD = "child"
    DISABLED = "disabled"
    INFANT = "infant"
    MILITARY = "military"
    SENIOR = "senior"
    STUDENT = "student"
    UNWAGED = "unwaged"
    YOUTH = "youth"
    STANDARD = "standard"



from collections import OrderedDict
from pyticketswitch.callout import Callout, Integration


class TestCallout:

    def test_from_api_data(self):

        data = {
            'bundle_source_code': 'ext_test0',
            'bundle_source_desc': 'External Test Backend 0',
            'bundle_total_cost': 79.5,
            'callout_destination_url': 'https://foobar.com',
            'callout_integration_data': {
                'foo': False,
                'bar': True,
                'lol': 'beans',
            },
            'callout_parameters': {
                'fours': 4,
                'twos': 222,
                'threes': 33,
                'ones': 1,
            },
            'callout_parameters_order': ['ones', 'twos', 'threes', 'fours'],
            'callout_type': 'get',
            'currency_code': 'gbp',
            'debitor': {
                'debitor_type': 'dummy'
            },
            'return_token': 'abc123'
        }

        callout = Callout.from_api_data(data)

        assert callout.code == 'ext_test0'
        assert callout.description == 'External Test Backend 0'
        assert callout.total == 79.5
        assert callout.destination == 'https://foobar.com'
        assert callout.parameters == OrderedDict(
            [
                ('ones', 1),
                ('twos', 222),
                ('threes', 33),
                ('fours', 4),
            ]
        )
        assert callout.integration_data == {
            'foo': False,
            'bar': True,
            'lol': 'beans',
        }
        assert callout.debitor.type == 'dummy'
        assert callout.currency_code == 'gbp'
        assert callout.return_token == 'abc123'

    def test_from_api_data_raw_callout(self):

        data = {
            'bundle_source_code': 'ext_test0',
            'bundle_source_desc': 'External Test Backend 0',
            'bundle_total_cost': 79.5,
            'callout_destination_url': 'https://foobar.com',
            'callout_integration_data': {
                'foo': False,
                'bar': True,
                'lol': 'beans',
            },
            'callout_parameters': {
                'fours': 4,
                'twos': 222,
                'threes': 33,
                'ones': 1,
            },
            'callout_type': 'get',
            'currency_code': 'gbp',
            'debitor': {
                'debitor_type': 'dummy'
            },
            'return_token': 'abc123'
        }

        callout = Callout.from_api_data(data)

        assert callout.parameters['fours'] == 4
        assert callout.parameters['threes'] == 33
        assert callout.parameters['twos'] == 222
        assert callout.parameters['ones'] == 1


class TestIntegration:

    def test_from_api_data(self):

        data = {
            'debit_amount': 62.5,
            'debit_base_amount': 6250,
            'debit_currency': 'gbp',
            'debitor_specific_data': {
                'is_dummy_3d_secure': False,
            },
            'debitor_type': 'dummy',
        }

        integration = Integration.from_api_data(data)
        assert integration.type == 'dummy'
        assert integration.amount == 62.5
        assert integration.base_amount == 6250
        assert integration.currency == 'gbp'
        assert integration.data == {
            'is_dummy_3d_secure': False,
        }



import datetime
from pyticketswitch.month import Month


class TestMonth:

    def test_from_api_data(self):
        data = {
            'month': 'nov',
            'year': 2018,
            'month_dates_bitmask': 1065254912,
            'month_desc': 'November',
            'month_weekdays_bitmask': 63,
            'cost_range': {
                'range_currency_code': 'gbp',
                'valid_quantities': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                'max_surcharge': 39.45,
                'max_seatprice': 197.25,
                'min_surcharge': 5.4,
                'min_seatprice': 20.0,
                'no_singles_cost_range': {
                    'range_currency_code': 'gbp',
                    'valid_quantities': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                    'max_surcharge': 39.45,
                    'max_seatprice': 197.25,
                    'min_surcharge': 5.4,
                    'min_seatprice': 30.0
                },
            }
        }

        month = Month.from_api_data(data)

        assert month.month == 11
        assert month.year == 2018
        assert month.description == 'November'
        assert month._dates_bitmask == 1065254912
        assert month._weekday_bitmask == 63
        assert month.cost_range.min_seatprice == 20.00
        assert month.cost_range.currency == 'gbp'
        assert month.no_singles_cost_range.min_seatprice == 30.00
        assert month.cost_range.currency == 'gbp'

    def test_from_api_data_without_cost_range(self):
        data = {
            'month': 'nov',
            'year': 2018,
            'month_dates_bitmask': 1065254912,
            'month_desc': 'November',
            'month_weekdays_bitmask': 63
        }

        month = Month.from_api_data(data)

        assert month.month == 11
        assert month.year == 2018
        assert month.description == 'November'
        assert month._dates_bitmask == 1065254912
        assert month._weekday_bitmask == 63
        assert month.cost_range is None
        assert month.no_singles_cost_range is None

    def test_is_available(self):
        # 1065254912 == 111111011111101000000000000000
        assert Month(0, 0, dates_bitmask=1065254912).is_available(1) is False
        assert Month(0, 0, dates_bitmask=1065254912).is_available(15) is False
        assert Month(0, 0, dates_bitmask=1065254912).is_available(16) is True
        assert Month(0, 0, dates_bitmask=1065254912).is_available(17) is False
        assert Month(0, 0, dates_bitmask=1065254912).is_available(23) is True
        assert Month(0, 0, dates_bitmask=1065254912).is_available(24) is False

    def test_on_weekday(self):
        # 85 == 1010101, and we add one to the index so monday is 0 rather than
        # sunday like in the api
        assert Month(0, 0, weekday_bitmask=85).on_weekday(0) is False
        assert Month(0, 0, weekday_bitmask=85).on_weekday(1) is True
        assert Month(0, 0, weekday_bitmask=85).on_weekday(2) is False
        assert Month(0, 0, weekday_bitmask=85).on_weekday(3) is True

    def test_start_date(self):
        # 1065254912 == 111111011111101000000000000000, and we add one to the index so monday is 0 rather than
        # sunday like in the api
        sixteenth_of_january = datetime.date(2020, 1, 16)
        assert Month(1, 2020, dates_bitmask=1065254912).start_date() == sixteenth_of_january

    def test_start_date_with_none(self):
        assert Month(1, 2020, dates_bitmask=None).start_date() is None

    def test_end_date(self):
        # 2130509824 == 1111110111111010000000000000000
        thirty_first_of_march = datetime.date(2020, 3, 31)
        assert Month(3, 2020, dates_bitmask=2130509824).end_date() == thirty_first_of_march

    def test_end_date_with_none(self):
        assert Month(1, 2020, dates_bitmask=None).end_date() is None



from decimal import Decimal
from pyticketswitch.bundle import Bundle
from pyticketswitch.event import Event
from pyticketswitch.order import Order
from pyticketswitch.debitor import Debitor
from pyticketswitch.purchase_result import PurchaseResult


class TestBundle(object):
    def test_from_api_data(self):
        data = {
            'bundle_order_count': 1,
            'bundle_source_code': 'ext_test0',
            'bundle_source_desc': 'External Test Backend 0',
            'bundle_total_cost': 54.7,
            'bundle_total_seatprice': 51,
            'bundle_total_send_cost': 1.5,
            'bundle_total_surcharge': 2.2,
            'bundle_send_cost_tax_component': 0.95,
            'bundle_total_combined_tax_component': 35.0,
            'bundle_total_surcharge_tax_sub_component': 0.0,
            'currency_code': 'gbp',
            'order': [
                {'item_number': 1},
                {'item_number': 2},
            ],
            "debitor": {
                "debitor_type": "dummy"
            },
            "source_t_and_c": 'some legal stuff',
            "purchase_result": {
                "success": True,
            },
        }

        bundle = Bundle.from_api_data(data)

        assert bundle.source_code == 'ext_test0'
        assert bundle.currency_code == 'gbp'
        assert bundle.terms_and_conditions == 'some legal stuff'

        assert len(bundle.orders) == 2
        assert bundle.orders[0].item == 1
        assert bundle.orders[1].item == 2
        assert bundle.description == 'External Test Backend 0'

        assert isinstance(bundle.total_seatprice, (int, float))
        assert bundle.total_seatprice == 51.0

        assert isinstance(bundle.total_surcharge, (int, float))
        assert bundle.total_surcharge == 2.2

        assert isinstance(bundle.total_send_cost, (int, float))
        assert bundle.total_send_cost == 1.5

        assert isinstance(bundle.total, (int, float))
        assert bundle.total == 54.7
        
        assert isinstance(bundle.send_cost_tax_component, (int, float))
        assert bundle.send_cost_tax_component == 0.95

        assert isinstance(bundle.total_combined_tax_component, (int, float))
        assert bundle.total_combined_tax_component == 35.0

        assert isinstance(bundle.total_surcharge_tax_sub_component, (int, float))
        assert bundle.total_surcharge_tax_sub_component == 0.0

        assert isinstance(bundle.debitor, Debitor)
        assert bundle.debitor.type == 'dummy'

    def test_from_api_data_with_decimals(self):
        data = {
            'bundle_order_count': 1,
            'bundle_source_code': 'ext_test0',
            'bundle_source_desc': 'External Test Backend 0',
            'bundle_total_cost': Decimal('54.7'),
            'bundle_total_seatprice': Decimal('51'),
            'bundle_total_send_cost': Decimal('1.5'),
            'bundle_total_surcharge': Decimal('2.2'),
            "bundle_send_cost_tax_component": Decimal('0.95'),
            'bundle_total_combined_tax_component': Decimal('35.0'),
            'bundle_total_surcharge_tax_sub_component': Decimal('0.0'),
            'currency_code': 'gbp',
            'order': [
                {'item_number': 1},
                {'item_number': 2},
            ],
            "debitor": {
                "debitor_type": "dummy"
            },
            "source_t_and_c": 'some legal stuff',
            "purchase_result": {
                "success": True,
            },
        }

        bundle = Bundle.from_api_data(data)

        assert bundle.source_code == 'ext_test0'
        assert bundle.currency_code == 'gbp'
        assert bundle.terms_and_conditions == 'some legal stuff'

        assert len(bundle.orders) == 2
        assert bundle.orders[0].item == 1
        assert bundle.orders[1].item == 2
        assert bundle.description == 'External Test Backend 0'

        assert isinstance(bundle.total_seatprice, Decimal)
        assert bundle.total_seatprice == Decimal('51.0')

        assert isinstance(bundle.total_surcharge, Decimal)
        assert bundle.total_surcharge == Decimal('2.2')

        assert isinstance(bundle.total_send_cost, Decimal)
        assert bundle.total_send_cost == Decimal('1.5')

        assert isinstance(bundle.total, Decimal)
        assert bundle.total == Decimal('54.7')

        assert isinstance(bundle.send_cost_tax_component, Decimal)
        assert bundle.send_cost_tax_component == Decimal('0.95')

        assert isinstance(bundle.total_combined_tax_component, Decimal)
        assert bundle.total_combined_tax_component == Decimal('35.0')

        assert isinstance(bundle.total_surcharge_tax_sub_component, Decimal)
        assert bundle.total_surcharge_tax_sub_component == Decimal('0.0')

        assert isinstance(bundle.debitor, Debitor)
        assert bundle.debitor.type == 'dummy'

    def test_get_events(self):

        event_one = Event(id_='abc123')
        event_two = Event(id_='def456')

        bundle = Bundle(
            'tests',
            orders=[
                Order(item=1, event=event_one),
                Order(item=2, event=event_two),
            ]
        )

        events = bundle.get_events()

        assert events == [event_one, event_two]

    def test_get_events_with_no_orders(self):

        bundle = Bundle('tests', orders=None)

        events = bundle.get_events()

        assert events == []

    def test_get_event_ids(self):
        event_one = Event(id_='abc123')
        event_two = Event(id_='def456')

        bundle = Bundle(
            'tests',
            orders=[
                Order(item=1, event=event_one),
                Order(item=2, event=event_two),
            ]
        )

        events = bundle.get_event_ids()

        assert events == {'abc123', 'def456'}

    def test_is_purchased(self):
        pr = PurchaseResult(success=True)

        bundle_orders = [Order(item=1, event=Event(id_='TEST'))]

        bundle = Bundle(
            'tests',
            orders=bundle_orders,
            purchase_result=pr,
        )

        no_pr_bundle = Bundle(
            'tests',
            orders=bundle_orders,
        )

        assert bundle.is_purchased()
        assert not no_pr_bundle.is_purchased()

    def test_repr(self):

        bundle = Bundle('tests')

        assert repr(bundle) == '<Bundle tests>'



from pyticketswitch.media import Media


class TestMedia:

    def test_from_api_data_secure(self):
        data = {
            'caption_html': '<BR>',
            'name': 'landscape',
            'secure_complete_url': 'https://testurl.com',
            'supports_http': True,
            'caption': 'caption',
            'host': 'd1wx4w35ubmdix.cloudfront.net',
            'supports_https': True,
            'path': '/shared/event_media/cropper_cloud/24/249fdb60b56ef4217a6049a1ed770f13552bca9c.jpg',
            'insecure_complete_url': 'http://testurl.com',
        }

        media = Media.from_api_data(data)
        assert media.url == 'https://testurl.com'
        assert media.caption_html == '<BR>'
        assert media.caption == 'caption'
        assert media.name == 'landscape'
        assert media.secure is True

    def test_from_api_data_insecure(self):
        data = {
            'caption_html': '<BR>',
            'name': 'landscape',
            'secure_complete_url': '',
            'supports_http': False,
            'caption': 'caption',
            'host': 'd1wx4w35ubmdix.cloudfront.net',
            'supports_https': True,
            'path': '/shared/event_media/cropper_cloud/24/249fdb60b56ef4217a6049a1ed770f13552bca9c.jpg',
            'insecure_complete_url': 'http://testurl.com',
        }

        media = Media.from_api_data(data)
        assert media.url == 'http://testurl.com'
        assert media.secure is False



from pyticketswitch.address import Address
from pyticketswitch.reservation import Reservation
from pyticketswitch.trolley import Trolley


class TestReservation:

    def test_from_api_data(self):

        data = {
            'allowed_countries': {
                'ad': 'Andorra',
                'ae': 'United Arab Emirates',
            },
            'can_edit_address': True,
            'needs_agent_reference': True,
            'needs_email_address': True,
            'needs_payment_card': True,
            'prefilled_address': {
                'address_line_one': 'Metro Building',
                'address_line_two': '1 Butterwick',
                'country_code': 'uk',
                'county': 'United Kingdom',
                'email_address': 'lol@beans.com',
                'home_phone': '020810101010101',
                'postcode': 'W6 8DL',
                'town': 'Hammersmith, London',
                'work_phone': '020801010101010'
            },
            'reserved_trolley': {
                'bundle': [
                    {
                        'bundle_source_code': 'ext_test0',
                        'order': [{'item_number': 1}],
                    }
                ],
                'random_index': '5d3928de-c923-11e6-b2f6-0025903268a0',
                'trolley_bundle_count': 1,
                'trolley_order_count': 1
            },
            'minutes_left_on_reserve': 15,
            "unreserved_orders": [{
                'item_number': 2,
                'reserve_failure_comment': 'failure reason',
            }],
        }

        reservation = Reservation.from_api_data(data)

        assert reservation.needs_payment_card is True
        assert reservation.needs_email_address is True
        assert reservation.needs_agent_reference is True
        assert reservation.can_edit_address is True

        assert isinstance(reservation.prefilled_address, Address)

        assert isinstance(reservation.trolley, Trolley)

        assert len(reservation.allowed_countries) == 2
        assert reservation.allowed_countries[0].code == 'ad'
        assert reservation.allowed_countries[1].code == 'ae'

        assert len(reservation.unreserved_orders) == 1
        assert reservation.minutes_left == 15
        assert reservation.unreserved_orders[0].reserve_failure_comment == 'failure reason'

    def test_from_api_data_with_unavailable_orders(self):
        data = {
            "input_contained_unavailable_order": True,
            "unreserved_orders": [],
        }

        reservation = Reservation.from_api_data(data)

        assert reservation.input_contained_unavailable_order is True



from pyticketswitch.content import Content


class TestContent:

    def test_from_api_data(self):

        data = {
            'name': 'name',
            'value': 'value',
            'value_html': '<p>value</p>',
        }
        content = Content.from_api_data(data)
        assert content.name == 'name'
        assert content.value == 'value'
        assert content.value_html == '<p>value</p>'



import decimal
import pytest
import json
import requests
from datetime import datetime
from mock import Mock
import pyticketswitch
from pyticketswitch.client import Client, POST, GET
from pyticketswitch import exceptions
from pyticketswitch.trolley import Trolley
from pyticketswitch.reservation import Reservation
from pyticketswitch.user import User
from pyticketswitch.customer import Customer
from pyticketswitch.payment_methods import CardDetails, RedirectionDetails
from pyticketswitch.status import Status
from pyticketswitch.callout import Callout



@pytest.fixture
def client():
    client = Client(user="bilbo", password="baggins", use_decimal=True)
    return client


@pytest.fixture
def fake_func():
    def wrapper(return_value):
        def fake(*args, **kwargs):
            return return_value
        return fake

    return wrapper


@pytest.fixture
def mock_make_request(client, monkeypatch):
    response = {'results': {}}
    mock_make_request = Mock(return_value=response)
    monkeypatch.setattr(client, 'make_request', mock_make_request)
    return mock_make_request


@pytest.fixture
def mock_make_request_for_events(client, monkeypatch):
    response = {'events_by_id': {}}
    mock_make_request = Mock(return_value=response)
    monkeypatch.setattr(client, 'make_request', mock_make_request)
    return mock_make_request


@pytest.fixture
def mock_make_request_for_performances(client, monkeypatch):
    response = {'performances_by_id': {}}
    mock_make_request = Mock(return_value=response)
    monkeypatch.setattr(client, 'make_request', mock_make_request)
    return mock_make_request


@pytest.fixture
def mock_make_request_for_availability(client, monkeypatch):
    response = {'availability': {}}
    mock_make_request = Mock(return_value=response)
    monkeypatch.setattr(client, 'make_request', mock_make_request)
    return mock_make_request


@pytest.fixture
def mock_make_request_for_trolley(client, monkeypatch):
    response = {'trolley_token': 'ABC123'}
    mock_make_request = Mock(return_value=response)
    monkeypatch.setattr(client, 'make_request', mock_make_request)
    return mock_make_request


class FakeResponse(object):

    def __init__(self, status_code=200, json=None):
        self.status_code = status_code
        self._json = json

    def json(self, **kwargs):
        return self._json

    @property
    def content(self):
        return json.dumps(self._json)


class FakeResponseRaisesValueError(FakeResponse):

    def json(self, **kwargs):
        raise ValueError("ERROR")


class TestClient:

    @pytest.mark.integration
    def test_get_url(self, client):
        url = client.get_url('events.v1')
        assert url == 'https://api.ticketswitch.com/f13/events.v1/'

    @pytest.mark.integration
    def test_make_request(self, client, monkeypatch):
        fake_response = FakeResponse(status_code=200, json={"lol": "beans"})
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))

        params = {
            'foo': 'bar',
        }
        client.language='en-GB'
        response = client.make_request('events.v1', params)
        assert response == {'lol': 'beans'}
        fake_get.assert_called_with(
            'https://api.ticketswitch.com/f13/events.v1/',
            auth=(b'bilbo', b'baggins'),
            params={
                'foo': 'bar',
            },
            headers={
                'Accept-Language': 'en-GB',
                'User-Agent': 'pyticketswitch {}'.format(pyticketswitch.__version__),
            },
            timeout=None
        )

    @pytest.mark.integration
    def test_make_request_with_timeout(self, client, monkeypatch):
        fake_response = FakeResponse(status_code=200, json={"lol": "beans"})
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))

        params = {
            'foo': 'bar',
        }
        client.language='en-GB'
        response = client.make_request('events.v1', params, timeout=15)
        assert response == {'lol': 'beans'}
        fake_get.assert_called_with(
            'https://api.ticketswitch.com/f13/events.v1/',
            auth=(b'bilbo', b'baggins'),
            params={
                'foo': 'bar',
            },
            headers={
                'Accept-Language': 'en-GB',
                'User-Agent': 'pyticketswitch {}'.format(pyticketswitch.__version__),
            },
            timeout=15
        )

    @pytest.mark.integration
    def test_make_request_with_post(self, client, monkeypatch):
        fake_response = FakeResponse(status_code=200, json={"lol": "beans"})
        fake_post = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.post = fake_post
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))

        params = {
            'foo': 'bar',
        }
        client.language='en-GB'
        response = client.make_request('events.v1', params, method=POST)
        assert response == {'lol': 'beans'}
        fake_post.assert_called_with(
            'https://api.ticketswitch.com/f13/events.v1/',
            auth=(b'bilbo', b'baggins'),
            data={
                'foo': 'bar',
            },
            headers={
                'Accept-Language': 'en-GB',
                'User-Agent': 'pyticketswitch {}'.format(pyticketswitch.__version__),
            },
            timeout=None
        )

    def test_make_request_with_subuser(self, monkeypatch):
        client = Client(user="beatles", password="lovemedo",
                        sub_user="ringo", use_decimal=True)
        fake_response = FakeResponse(status_code=200, json={"lol": "beans"})
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))

        params = {
            'foo': 'bar',
        }
        client.language='en-GB'
        response = client.make_request('events.v1', params)
        assert response == {'lol': 'beans'}
        fake_get.assert_called_with(
            'https://api.ticketswitch.com/f13/events.v1/',
            auth=(b'beatles', b'lovemedo'),
            params={
                'foo': 'bar',
                'sub_id': 'ringo',
            },
            headers={
                'Accept-Language': 'en-GB',
                'User-Agent': 'pyticketswitch {}'.format(pyticketswitch.__version__),
            },
            timeout=None
        )

    def test_make_request_with_tracking_id(self, monkeypatch):
        client = Client(user="user", password="pass",
                        tracking_id="xyz", use_decimal=True)
        fake_response = FakeResponse(status_code=200, json={"depro": "fundis"})
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))
        client.language='en-GB'
        response = client.make_request('events.v1', {})
        assert response
        fake_get.assert_called_with(
            'https://api.ticketswitch.com/f13/events.v1/',
            auth=(b'user', b'pass'),
            params={
                'tsw_session_track_id': 'xyz'
            },
            headers={
                'Accept-Language': 'en-GB',
                'User-Agent': 'pyticketswitch {}'.format(pyticketswitch.__version__),
            },
            timeout=None
        )

    def test_make_request_when_using_per_request_tracking_id(self, monkeypatch):
        client = Client(user="user", password="pass",
                        tracking_id="xyz", use_decimal=True)
        fake_response = FakeResponse(status_code=200, json={"depro": "fundis"})
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))
        client.language='en-GB'
        params = {}
        client.add_optional_kwargs(params, tracking_id="123")
        response = client.make_request('events.v1', params)

        assert response
        fake_get.assert_called_with(
            'https://api.ticketswitch.com/f13/events.v1/',
            auth=(b'user', b'pass'),
            params={
                'tsw_session_track_id': '123'
            },
            headers={
                'Accept-Language': 'en-GB',
                'User-Agent': 'pyticketswitch {}'.format(pyticketswitch.__version__),
            },
            timeout=None
        )

        client.add_optional_kwargs(params, tracking_id="456")
        fake_get.assert_called_with(
            'https://api.ticketswitch.com/f13/events.v1/',
            auth=(b'user', b'pass'),
            params={
                'tsw_session_track_id': '456'
            },
            headers={
                'Accept-Language': 'en-GB',
                'User-Agent': 'pyticketswitch {}'.format(pyticketswitch.__version__),
            },
            timeout=None
        )

    def test_make_request_bad_response_with_auth_error(self, client, monkeypatch):
        fake_response = FakeResponse(status_code=400, json={
            'error_code': 3,
            'error_desc': 'User authorisation failure',
        })
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))

        with pytest.raises(exceptions.APIError) as excinfo:
            client.make_request('test.v1', {})

        assert excinfo.value.msg == 'User authorisation failure'
        assert excinfo.value.code == 3
        assert excinfo.value.response is fake_response

    def test_make_request_bad_response_with_error(self, client, monkeypatch):
        fake_response = FakeResponse(status_code=400, json={
            'error_code': 8,
            'error_desc': 'price_band_code needs /pool or /alloc suffix',
        })
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))
        with pytest.raises(exceptions.APIError) as excinfo:
            client.make_request('trolley.v1', {})

        assert excinfo.value.msg == 'price_band_code needs /pool or /alloc suffix'
        assert excinfo.value.code == 8
        assert excinfo.value.response is fake_response

    def test_make_request_bad_response_without_error(self, client, monkeypatch):
        fake_response = FakeResponse(status_code=400, json={})
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))
        with pytest.raises(exceptions.InvalidResponseError):
            client.make_request('trolley.v1', {})

    def test_make_request_410_gone_response(self, client, monkeypatch):
        response_json = {'error_code': 8, 'error_desc': 'transaction failed'}
        fake_response = FakeResponse(status_code=410, json=response_json)
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))
        with pytest.raises(exceptions.CallbackGoneError):
            client.make_request('callback.v1', {})

    def test_make_request_no_contents_raises(self, client, monkeypatch):
        response_json = {'data': 'some data'}
        fake_response = FakeResponseRaisesValueError(status_code=200, json=response_json)
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))
        with pytest.raises(exceptions.InvalidResponseError):
            client.make_request('test.v1', {})

    def test_add_optional_kwargs_extra_info(self, client):
        params = {}
        client.add_optional_kwargs(params, extra_info=True)

        assert params == {'req_extra_info': True}

    def test_add_optional_kwargs_reviews(self, client):
        params = {}
        client.add_optional_kwargs(params, reviews=True)

        assert params == {'req_reviews': True}

    def test_add_optional_kwargs_media(self, client):
        params = {}
        client.add_optional_kwargs(params, media=True)

        assert params == {
            'req_media_triplet_one': True,
            'req_media_triplet_two': True,
            'req_media_triplet_three': True,
            'req_media_triplet_four': True,
            'req_media_triplet_five': True,
            'req_media_triplet_one_wide': True,
            'req_media_triplet_two_wide': True,
            'req_media_triplet_three_wide': True,
            'req_media_triplet_four_wide': True,
            'req_media_triplet_five_wide': True,
            'req_media_seating_plan': True,
            'req_media_square': True,
            'req_media_landscape': True,
            'req_media_marquee': True,
            'req_video_iframe': True,
            'req_media_supplier': True,
        }

    def test_add_optional_kwargs_cost_range(self, client):
        params = {}
        client.add_optional_kwargs(params, cost_range=True)

        assert params == {'req_cost_range': True}

    def test_add_optional_kwargs_best_value_offer(self, client):
        params = {}
        client.add_optional_kwargs(params, best_value_offer=True)

        assert params == {
            'req_cost_range': True,
            'req_cost_range_best_value_offer': True,
        }

    def test_add_optional_kwargs_max_saving_offer(self, client):
        params = {}
        client.add_optional_kwargs(params, max_saving_offer=True)

        assert params == {
            'req_cost_range': True,
            'req_cost_range_max_saving_offer': True,
        }

    def test_add_optional_kwargs_min_cost_offer(self, client):
        params = {}
        client.add_optional_kwargs(params, min_cost_offer=True)

        assert params == {
            'req_cost_range': True,
            'req_cost_range_min_cost_offer': True,
        }

    def test_add_optional_kwargs_top_price_offer(self, client):
        params = {}
        client.add_optional_kwargs(params, top_price_offer=True)

        params == {
            'req_cost_range': True,
            'req_cost_range_top_price_offer': True,
        }

    def test_add_optional_kwargs_no_singles_data(self, client):
        params = {}
        client.add_optional_kwargs(params, no_singles_data=True)

        assert params == {
            'req_cost_range': True,
            'req_cost_range_no_singles_data': True,
        }

    def test_add_optional_kwargs_cost_range_details(self, client):
        params = {}
        client.add_optional_kwargs(params, cost_range_details=True)

        assert params == {
            'req_cost_range_details': True,
        }

    def test_add_optional_kwargs_avail_details(self, client):
        params = {}
        client.add_optional_kwargs(params, availability=True)

        params == {
            'req_avail_details': True,
        }

    def test_add_optional_kwargs_avail_details_with_perfs(self, client):
        params = {}
        client.add_optional_kwargs(params, availability_with_performances=True)

        params == {
            'req_avail_details_with_perfs': True,
        }

    def test_add_optional_kwargs_source_info(self, client):
        params = {}
        client.add_optional_kwargs(params, source_info=True)

        params == {
            'req_src_info': True,
        }

    def test_list_events(self, client, monkeypatch):
        response = {
            'results': {
                'event': [
                    {'event_id': 'ABC123'},
                    {'event_id': 'DEF456'},
                ],
                'paging_status': {
                    'total_unpaged_results': 10,
                },
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }
        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        events, meta = client.list_events()

        mock_make_request.assert_called_with('events.v1', {})

        assert len(events) == 2
        event_one, event_two = events

        assert event_one.id =='ABC123'
        assert event_two.id == 'DEF456'

        assert meta.total_results == 10
        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_list_events_with_keywords(self, client, mock_make_request):
        client.list_events(keywords=['awesome', 'stuff'])

        mock_make_request.assert_called_with('events.v1', {
            'keywords': 'awesome,stuff',
        })

    def test_list_events_with_start_date(self, client, mock_make_request):
        client.list_events(start_date=datetime(2016, 7, 23, 0, 7, 25))

        mock_make_request.assert_called_with('events.v1', {
            'date_range': '20160723:',
        })

    def test_list_events_with_end_date(self, client, mock_make_request):
        client.list_events(end_date=datetime(2016, 7, 23, 0, 7, 25))

        mock_make_request.assert_called_with('events.v1', {
            'date_range': ':20160723',
        })

    def test_list_events_with_start_and_end_date(self, client, mock_make_request):
        client.list_events(
            start_date=datetime(2015, 3, 11, 0, 9, 45),
            end_date=datetime(2016, 7, 23, 0, 7, 25)
        )

        mock_make_request.assert_called_with('events.v1', {
            'date_range': '20150311:20160723',
        })

    def test_list_events_country_code(self, client, mock_make_request):
        client.list_events(country_code='fj')

        mock_make_request.assert_called_with('events.v1', {
            'country_code': 'fj',
        })

    def test_list_events_city_code(self, client, mock_make_request):
        client.list_events(city_code='london-uk')

        mock_make_request.assert_called_with('events.v1', {
            'city_code': 'london-uk',
        })

    def test_list_events_geolocation(self, client, mock_make_request):
        client.list_events(
            latitude=51.52961137,
            longitude=-0.10601562,
            radius=10
        )

        mock_make_request.assert_called_with('events.v1', {
            'circle': '51.52961137:-0.10601562:10',
        })

    def test_list_events_invalid_geolocation(self, client):
        with pytest.raises(exceptions.InvalidGeoParameters):
            client.list_events(
                longitude=-0.10601562,
                radius=10
            )

        with pytest.raises(exceptions.InvalidGeoParameters):
            client.list_events(
                latitude=51.52961137,
                radius=10
            )

        with pytest.raises(exceptions.InvalidGeoParameters):
            client.list_events(
                latitude=51.52961137,
                longitude=-0.10601562,
            )

        with pytest.raises(exceptions.InvalidGeoParameters):
            client.list_events(
                radius=10
            )

    def test_list_events_include_dead(self, client, mock_make_request):
        client.list_events(include_dead=True)

        mock_make_request.assert_called_with('events.v1', {
            'include_dead': True,
        })

    def test_list_events_sort_order(self, client, mock_make_request):
        client.list_events(sort_order='foobar')

        mock_make_request.assert_called_with('events.v1', {
            'sort_order': 'foobar',
        })

    def test_list_events_pagination(self, client, mock_make_request):
        client.list_events(page=2, page_length=50)

        mock_make_request.assert_called_with('events.v1', {
            'page_no': 2,
            'page_len': 50,
        })

    def test_list_events_no_results(self, client, monkeypatch, fake_func):
        response = {}
        monkeypatch.setattr(client, 'make_request', fake_func(response))

        with pytest.raises(exceptions.InvalidResponseError):
            client.list_events()

    def test_list_events_misc_kwargs(self, client, mock_make_request):
        client.list_events(foobar='lolbeans')

        mock_make_request.assert_called_with('events.v1', {
            'foobar': 'lolbeans'
        })

    def test_get_events(self, client, monkeypatch):
        response = {
            'events_by_id': {
                'ABC123': {
                    'event': {'event_id': 'ABC123'},
                },
                'DEF456': {
                    'event': {'event_id': 'DEF456'},
                }
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        events, meta = client.get_events(['ABC123', 'DEF456'])

        mock_make_request.assert_called_with(
            'events_by_id.v1',
            {'event_id_list': 'ABC123,DEF456'},
        )

        event_one = events['ABC123']
        event_two = events['DEF456']

        assert event_one.id == 'ABC123'
        assert event_two.id == 'DEF456'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_get_events_event_list(self, client, mock_make_request_for_events):
        client.get_events(['6IF', '25DR', '3ENO'])

        mock_make_request_for_events.assert_called_with('events_by_id.v1', {
            'event_id_list': '6IF,25DR,3ENO',
        })

    def test_get_events_no_results(self, client, monkeypatch, fake_func):
        response = {}
        monkeypatch.setattr(client, 'make_request', fake_func(response))

        with pytest.raises(exceptions.InvalidResponseError):
            client.get_events(['6IF', '25DR'])

    def test_get_events_misc_kwargs(self, client, mock_make_request_for_events):
        client.get_events([], foobar='lolbeans')

        mock_make_request_for_events.assert_called_with('events_by_id.v1', {
            'foobar': 'lolbeans',
        })

    def test_get_events_with_upsell(self, client, mock_make_request_for_events):
        client.get_events(['6IF'], with_upsells=True)

        mock_make_request_for_events.assert_called_with('events_by_id.v1', {
            'event_id_list': '6IF', 'add_upsells': True,
        })

    def test_get_events_with_addons(self, client, mock_make_request_for_events):
        client.get_events(['ABC123'], with_addons=True)

        mock_make_request_for_events.assert_called_with('events_by_id.v1', {
            'event_id_list': 'ABC123', 'add_add_ons': True,
        })

    def test_get_event(self, client, monkeypatch):
        response = {
            'events_by_id': {
                'ABC123': {
                    'event': {'event_id': 'ABC123'},
                },
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        event, meta = client.get_event('ABC123')

        mock_make_request.assert_called_with(
            'events_by_id.v1',
            {'event_id_list': 'ABC123'},
        )
        assert event.id =='ABC123'
        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_get_months(self, client, monkeypatch):
        response = {
            'results': {
                'month': [
                    {'month': 'dec', 'year': 2016},
                    {'month': 'jan', 'year': 2017},
                    {'month': 'feb', 'year': 2017},
                ]
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        months = client.get_months('ABC123')

        mock_make_request.assert_called_with(
            'months.v1',
            {'event_id': 'ABC123'},
        )

        assert len(months) == 3

        assert months[0].month == 12
        assert months[0].year == 2016
        assert months[1].month == 1
        assert months[1].year == 2017
        assert months[2].month == 2
        assert months[2].year == 2017

    def test_get_months_no_results(self, client, monkeypatch, fake_func):
        response = {}
        monkeypatch.setattr(client, 'make_request', fake_func(response))

        with pytest.raises(exceptions.InvalidResponseError):
            client.get_months('6IF')

    def test_get_months_misc_kwargs(self, client, mock_make_request):
        client.get_months('6IF', foobar='lolbeans')

        mock_make_request.assert_called_with('months.v1', {
            'event_id': '6IF',
            'foobar': 'lolbeans'
        })

    def test_list_performances_no_results(self, client, monkeypatch, fake_func):
        response = {}
        monkeypatch.setattr(client, 'make_request', fake_func(response))

        with pytest.raises(exceptions.InvalidResponseError):
            client.list_performances('6IF')

    def test_list_performances(self, client, monkeypatch):
        response = {
            'results': {
                'has_perf_names': False,
                'events_by_id': {
                    'ABC123': {'event': {'event_id': 'ABC123'}},
                },
                'performance': [
                    {'perf_id': 'ABC123-1', 'event_id': 'ABC123'},
                    {'perf_id': 'ABC123-2', 'event_id': 'ABC123'},
                    {'perf_id': 'ABC123-3', 'event_id': 'ABC123'},
                ],
                'paging_status': {
                    'total_unpaged_results': 10,
                },
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        performances, meta = client.list_performances('ABC123')

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
        })

        assert len(performances) == 3

        performance_one, performance_two, performance_three = performances

        assert performance_one.id == 'ABC123-1'
        assert performance_two.id == 'ABC123-2'
        assert performance_three.id == 'ABC123-3'

        assert performance_one.event_id == 'ABC123'
        assert performance_two.event_id == 'ABC123'
        assert performance_three.event_id == 'ABC123'

        assert meta.has_names is False
        assert meta.total_results == 10

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_list_performances_cost_range(self, client, mock_make_request):
        client.list_performances('ABC123', cost_range=True)

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'req_cost_range': True
        })

    def test_list_performances_best_value_offer(self, client, mock_make_request):
        client.list_performances('ABC123', best_value_offer=True)

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'req_cost_range': True,
            'req_cost_range_best_value_offer': True
        })

    def test_list_performances_max_saving_offer(self, client, mock_make_request):
        client.list_performances('ABC123', max_saving_offer=True)

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'req_cost_range': True,
            'req_cost_range_max_saving_offer': True
        })

    def test_list_performances_min_cost_offer(self, client, mock_make_request):
        client.list_performances('ABC123', min_cost_offer=True)

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'req_cost_range': True,
            'req_cost_range_min_cost_offer': True
        })

    def test_list_performances_top_price_offer(self, client, mock_make_request):
        client.list_performances('ABC123', top_price_offer=True)

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'req_cost_range': True,
            'req_cost_range_top_price_offer': True
        })

    def test_list_performances_no_singles_data(self, client, mock_make_request):
        client.list_performances('ABC123', no_singles_data=True)

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'req_cost_range': True,
            'req_cost_range_no_singles_data': True
        })

    def test_list_performances_availability(self, client, mock_make_request):
        client.list_performances('ABC123', availability=True)

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'req_avail_details': True
        })

    def test_list_performances_pagination(self, client, mock_make_request):
        client.list_performances(
            'ABC123',
            availability=True,
            page=3,
            page_length=20,
        )

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'req_avail_details': True,
            'page_no': 3,
            'page_len': 20,
        })

    def test_list_performances_with_start_date(self, client, mock_make_request):
        client.list_performances(
            'ABC123',
            start_date=datetime(2016, 7, 23, 0, 7, 25)
        )

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'date_range': '20160723:',
        })

    def test_list_performancess_with_end_date(self, client, mock_make_request):
        client.list_performances(
            'ABC123',
            end_date=datetime(2016, 7, 23, 0, 7, 25)
        )

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'date_range': ':20160723',
        })

    def test_list_performances_with_start_and_end_date(self, client, mock_make_request):
        client.list_performances(
            'ABC123',
            start_date=datetime(2015, 3, 11, 0, 9, 45),
            end_date=datetime(2016, 7, 23, 0, 7, 25)
        )

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'date_range': '20150311:20160723',
        })

    def test_list_performances_misc_kwargs(self, client, mock_make_request):
        client.list_performances('ABC123', foobar='lolbeans')

        mock_make_request.assert_called_with('performances.v1', {
            'event_id': 'ABC123',
            'foobar': 'lolbeans',
        })

    def test_get_performances(self, client, monkeypatch):
        response = {
            'performances_by_id': {
                'ABC123-1': {
                    'perf_id': 'ABC123-1',
                    'event_id': 'ABC123',
                },
                'DEF456-2': {
                    'perf_id': 'DEF456-2',
                    'event_id': 'DEF456',
                }
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        performances, meta = client.get_performances(['ABC123-1', 'DEF456-2'])

        mock_make_request.assert_called_with('performances_by_id.v1', {
            'perf_id_list': 'ABC123-1,DEF456-2',
        })

        performance_one = performances['ABC123-1']
        performance_two = performances['DEF456-2']

        assert performance_one.id == 'ABC123-1'
        assert performance_two.id == 'DEF456-2'

        assert performance_one.event_id == 'ABC123'
        assert performance_two.event_id == 'DEF456'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_get_performances_no_performances(self, client, monkeypatch, fake_func):
        response = {}
        monkeypatch.setattr(client, 'make_request', fake_func(response))

        with pytest.raises(exceptions.InvalidResponseError):
            client.get_performances(['6IF-1', '6IF-2'])

    def test_get_performances_misc_kwargs(self, client, mock_make_request_for_performances):
        client.get_performances(['6IF-1', '25DR-2'], foobar='lolbeans')

        mock_make_request_for_performances.assert_called_with('performances_by_id.v1', {
            'perf_id_list': '6IF-1,25DR-2',
            'foobar': 'lolbeans',
        })

    def test_get_performance(self, client, monkeypatch):
        response = {
            'performances_by_id': {
                'ABC123-1': {
                    'perf_id': 'ABC123-1',
                    'event_id': 'ABC123',
                },
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        performance, meta = client.get_performance('ABC123-1')

        mock_make_request.assert_called_with(
            'performances_by_id.v1',
            {'perf_id_list': 'ABC123-1'},
        )
        assert performance.id =='ABC123-1'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_get_availability(self, client, monkeypatch):
        response = {
            'availability': {
                'ticket_type': [
                    {
                        'ticket_type_code': 'CIRCLE',
                        'price_band': [
                            {
                                'price_band_code': 'A',
                            },
                            {
                                'price_band_code': 'B',
                                'allows_leaving_single_seats': 'if_necessary',
                            },
                        ]
                    },
                    {
                        'ticket_type_code': 'STALLS',
                        'price_band': [
                            {
                                'price_band_code': 'C',
                                'allows_leaving_single_seats': 'always',
                            },
                            {
                                'price_band_code': 'D',
                                'allows_leaving_single_seats': 'never',
                            },
                        ]
                    }
                ]
            },

            'backend_is_broken': False,
            'backend_is_down': False,
            'backend_throttle_failed': False,
            'contiguous_seat_selection_only': True,
            'must_select_whole_seat_block': True,
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            },
            'valid_quantities': [2, 3, 4, 5, 6, 7],
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        availability, meta = client.get_availability('ABC123-1')

        mock_make_request.assert_called_with('availability.v1', {
            'perf_id': 'ABC123-1',
        })

        assert meta.contiguous_seat_selection_only is True
        assert meta.must_select_whole_seat_block is True
        assert meta.default_currency_code == 'gbp'
        assert meta.valid_quantities == [2, 3, 4, 5, 6, 7]

        assert len(availability) == 2

        ticket_type_one = availability[0]
        assert ticket_type_one.code == 'CIRCLE'

        assert len(ticket_type_one.price_bands) == 2

        price_band_one = ticket_type_one.price_bands[0]
        assert price_band_one.code == 'A'

        price_band_two = ticket_type_one.price_bands[1]
        assert price_band_two.code == 'B'
        assert price_band_two.allows_leaving_single_seats == 'if_necessary'

        ticket_type_two = availability[1]
        assert ticket_type_two.code == 'STALLS'

        assert len(ticket_type_two.price_bands) == 2

        price_band_three = ticket_type_two.price_bands[0]
        assert price_band_three.code == 'C'
        assert price_band_three.allows_leaving_single_seats == 'always'

        price_band_four = ticket_type_two.price_bands[1]
        assert price_band_four.code == 'D'
        assert price_band_four.allows_leaving_single_seats == 'never'

    def test_get_availability_with_number_of_seats(self, client, mock_make_request_for_availability):
        client.get_availability('6IF-1', number_of_seats=2)

        mock_make_request_for_availability.assert_called_with('availability.v1', {
            'perf_id': '6IF-1',
            'no_of_seats': 2,
        })

    def test_get_availability_with_discounts(self, client, mock_make_request_for_availability):
        client.get_availability('6IF-1', discounts=True)

        mock_make_request_for_availability.assert_called_with('availability.v1', {
            'perf_id': '6IF-1',
            'add_discounts': True
        })

    def test_get_availability_with_example_seats(self, client, mock_make_request_for_availability):
        client.get_availability('6IF-1', example_seats=True)

        mock_make_request_for_availability.assert_called_with('availability.v1', {
            'perf_id': '6IF-1',
            'add_example_seats': True
        })

    def test_get_availability_with_seat_blocks(self, client, mock_make_request_for_availability):
        client.get_availability('6IF-1', seat_blocks=True)

        mock_make_request_for_availability.assert_called_with('availability.v1', {
            'perf_id': '6IF-1',
            'add_seat_blocks': True
        })

    def test_get_availability_with_user_commission(self, client, mock_make_request_for_availability):
        client.get_availability('6IF-1', user_commission=True)

        mock_make_request_for_availability.assert_called_with('availability.v1', {
            'perf_id': '6IF-1',
            'req_predicted_commission': True,
        })

    def test_get_availability_no_availability(self, client, monkeypatch):
        response = {
            'backend_is_broken': False,
            'backend_is_down': False,
            'backend_throttle_failed': False,
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        with pytest.raises(exceptions.InvalidResponseError):
            _, _ = client.get_availability('ABC123-1')

    def test_get_send_methods(self, client, monkeypatch):
        response = {
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            },
            'send_methods': {
                'send_method': [
                    {'send_code': 'COBO'},
                    {'send_code': 'POST'}
                ]
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        send_methods, meta = client.get_send_methods('ABC123-1')

        mock_make_request.assert_called_with('send_methods.v1', {
            'perf_id': 'ABC123-1',
        })

        assert len(send_methods) == 2
        assert send_methods[0].code == 'COBO'
        assert send_methods[1].code == 'POST'

        assert meta.get_currency().code == 'gbp'

    def test_get_send_methods_bad_data(self, client, monkeypatch):
        mock_make_request = Mock(return_value={})
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        with pytest.raises(exceptions.InvalidResponseError):
            client.get_send_methods('ABC123-1')

    def test_get_discounts(self, client, monkeypatch):
        response = {
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            },
            'discounts': {
                'discount': [
                    {'discount_code': 'ADULT'},
                    {'discount_code': 'CHILD'}
                ]
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        discounts, meta = client.get_discounts('ABC123-1', 'STALLS', 'A/pool',
                                               an_optional_kwarg='kwarg_value')

        mock_make_request.assert_called_with('discounts.v1', {
            'perf_id': 'ABC123-1',
            'ticket_type_code': 'STALLS',
            'price_band_code': 'A/pool',
            'req_predicted_commission': False,
            'an_optional_kwarg': 'kwarg_value',
        })

        assert len(discounts) == 2
        assert discounts[0].code == 'ADULT'
        assert discounts[1].code == 'CHILD'

        assert meta.get_currency().code == 'gbp'

    def test_get_discounts_bad_data(self, client, monkeypatch):
        mock_make_request = Mock(return_value={})
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        with pytest.raises(exceptions.InvalidResponseError):
            client.get_discounts('ABC123-1', 'STALLS', 'A/pool')

    def test_trolley_params_with_trolley_token(self, client):
        params = client._trolley_params(token='DEF456')

        assert params == {'trolley_token': 'DEF456'}

    def test_trolley_params_with_performance_id(self, client):
        params = client._trolley_params(performance_id='6IF-A8B')

        assert params == {'perf_id': '6IF-A8B'}

    def test_trolley_params_with_number_of_seats(self, client):
        params = client._trolley_params(number_of_seats=3)

        assert params == {'no_of_seats': 3}

    def test_trolley_params_with_ticket_type_code(self, client):
        params = client._trolley_params(ticket_type_code='STALLS')

        assert params == {'ticket_type_code': 'STALLS'}

    def test_trolley_params_with_price_band_code(self, client):
        params = client._trolley_params(price_band_code='A')

        assert params == {
            'price_band_code': 'A'
        }

    def test_trolley_params_with_item_numbers_to_remove(self, client):
        params = client._trolley_params(item_numbers_to_remove=[1, 2, 3], token='ABC123')

        assert params == {
            'trolley_token': 'ABC123',
            'remove_items_list': '1,2,3'
        }

    def test_trolley_params_with_item_numbers_to_remove_with_no_token(self, client):
        with pytest.raises(exceptions.InvalidParametersError):
            client._trolley_params(item_numbers_to_remove=[1, 2, 3])

    def test_trolley_params_with_seats(self, client):
        params = client._trolley_params(seats=['A12', 'B13', 'C14'])

        assert params == {
            'seat0': 'A12',
            'seat1': 'B13',
            'seat2': 'C14',
        }

    def test_trolley_params_with_discounts(self, client):
        params = client._trolley_params(discounts=['ADULT', 'CHILD', 'SENIOR'])

        assert params == {
            'disc0': 'ADULT',
            'disc1': 'CHILD',
            'disc2': 'SENIOR',
        }

    def test_trolley_params_with_send_codes(self, client):
        params = client._trolley_params(send_codes={'nimax': 'POST', 'see': 'COBO'})

        assert params == {
            'nimax_send_code': 'POST',
            'see_send_code': 'COBO'
        }

    def test_trolley_params_with_invalid_send_codes(self, client):
        with pytest.raises(exceptions.InvalidParametersError):
            client._trolley_params(send_codes=['POST', 'COBO'])

    def test_get_trolley(self, client, monkeypatch):
        response = {
            'trolley_contents': {},
            'trolley_token': 'DEF456',
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        trolley, meta = client.get_trolley()

        mock_make_request.assert_called_with('trolley.v1', {})

        assert isinstance(trolley, Trolley)
        assert trolley.token == 'DEF456'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_get_trolley_with_unavailable_order(self, client, monkeypatch):
        """
        This test is to check that an unavailable order doesn't raise
        any exceptions unless `raise_on_unavailable_order` is set to true
        """
        response = {
            'trolley_contents': {},
            'trolley_token': 'DEF456',
            'currency_code': 'gbp',
            'input_contained_unavailable_order': True,
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        # this should not raise any exceptions
        client.get_trolley()

        # but this should
        with pytest.raises(exceptions.OrderUnavailableError):
            client.get_trolley(raise_on_unavailable_order=True)


    def test_get_upsells(self, client, monkeypatch):
        # fakes
        response = {
            'results': {
                'event': [
                    {'event_id': 'GHI789'},
                    {'event_id': 'JKL012'},
                ],
                'paging_status': {
                    'total_unpaged_results': 2,
                },
            },
        }
        mock_make_request = Mock(return_value=response)

        monkeypatch.setattr(client, 'make_request', mock_make_request)

        # action
        (upsell_events, upsell_meta) = client.get_upsells(token="foobar")

        # results
        mock_make_request.assert_called_with('upsells.v1', {
            'trolley_token': 'foobar',
        })

        assert len(upsell_events) == 2
        event_one, event_two = upsell_events
        assert event_one.id == 'GHI789'
        assert event_two.id == 'JKL012'
        assert upsell_meta.total_results == 2

    def test_get_addons(self, client, monkeypatch):
        # fakes
        response = {
            'results': {
                'event': [
                    {'event_id': 'ABC123'},
                    {'event_id': 'DEF456'},
                ],
                'paging_status': {
                    'total_unpaged_results': 10,
                },
            },
        }
        mock_make_request = Mock(return_value=response)

        monkeypatch.setattr(client, 'make_request', mock_make_request)

        # action
        addon_events, addon_meta = client.get_addons(token="foobar")

        # results
        mock_make_request.assert_called_with('add_ons.v1', {
            'trolley_token': 'foobar',
        })

        assert len(addon_events) == 2
        event_one, event_two = addon_events
        assert event_one.id =='ABC123'
        assert event_two.id == 'DEF456'

        assert addon_meta.total_results == 10

    def test_make_reservation(self, client, monkeypatch):
        response = {
            'reserved_trolley': {
                'transaction_uuid': 'DEF456'
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        reservation, meta = client.make_reservation()

        mock_make_request.assert_called_with('reserve.v1', {}, method=POST)

        assert isinstance(reservation, Reservation)
        assert reservation.trolley.transaction_uuid == 'DEF456'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_get_reservation(self, client, monkeypatch):
        transaction_uuid = 'DEF456'
        response = {
            'reserved_trolley': {
                'transaction_uuid': transaction_uuid
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        reservation, meta = client.get_reservation(transaction_uuid)

        mock_make_request.assert_called_with('reserve_page_archive.v1', {
            "transaction_uuid": transaction_uuid
        }, method=GET)

        assert isinstance(reservation, Reservation)
        assert reservation.trolley.transaction_uuid == transaction_uuid

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_make_reservation_with_unavailable_order(self, client, monkeypatch):
        """
        This test is to check that an unavailable order doesn't raise
        any exceptions unless `raise_on_unavailable_order` is set to true
        """
        data = {
            "input_contained_unavailable_order": True,
            "unreserved_orders": [],
        }

        mock_make_request = Mock(return_value=data)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        # this should not raise any exceptions
        client.make_reservation()

        # but this should
        with pytest.raises(exceptions.OrderUnavailableError):
            client.make_reservation(raise_on_unavailable_order=True)

    def test_make_reservation_with_unavailable_order_but_successfull_reservation(self, client, monkeypatch):
        """
        This checks that when we raise an exception on unavailable order, but
        other parts of the trolley are successfully reserved, that we don't
        lose the transaction_uuid
        """
        data = {
            "input_contained_unavailable_order": True,
            'reserved_trolley': {
                'transaction_uuid': 'DEF456'
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=data)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        # but this should
        with pytest.raises(exceptions.OrderUnavailableError) as excinfo:
            client.make_reservation(raise_on_unavailable_order=True)

        exception = excinfo.value
        assert exception.reservation
        assert exception.reservation.trolley.transaction_uuid == 'DEF456'
        assert exception.meta.default_currency_code == 'gbp'

    def test_get_reservation_with_unavailable_order_but_successful_reservation(self, client, monkeypatch):
        """
        This checks that when we raise an exception on unavailable order, but
        other parts of the trolley are successfully reserved, that we don't
        lose the transaction_uuid
        """
        transaction_uuid = 'DEF456'
        data = {
            "input_contained_unavailable_order": True,
            'reserved_trolley': {
                'transaction_uuid': transaction_uuid
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=data)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        # but this should
        with pytest.raises(exceptions.OrderUnavailableError) as excinfo:
            client.get_reservation(transaction_uuid, raise_on_unavailable_order=True)

        exception = excinfo.value
        assert exception.reservation
        assert exception.reservation.trolley.transaction_uuid == transaction_uuid
        assert exception.meta.default_currency_code == 'gbp'


    def test_get_status(self, client, monkeypatch):
        response = {
            'trolley_contents': {
                'transaction_uuid': 'DEF456'
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        status, meta = client.get_status(
            transaction_uuid='DEF456',
            customer=True,
            external_sale_page=True,
        )

        mock_make_request.assert_called_with('status.v1', {
            'transaction_uuid': 'DEF456',
            'add_customer': True,
            'add_external_sale_page': True,
        })

        assert isinstance(status, Status)
        assert status.trolley.transaction_uuid == 'DEF456'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_get_status_with_trans(self, client, monkeypatch):
        response = {
            'trolley_contents': {
                'transaction_id': 'DEF456'
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        status, meta = client.get_status(
            transaction_id='DEF456',
            customer=True,
            external_sale_page=True,
        )

        mock_make_request.assert_called_with('trans_id_status.v1', {
            'transaction_id': 'DEF456',
            'add_customer': True,
            'add_external_sale_page': True,
        })

        assert isinstance(status, Status)
        assert status.trolley.transaction_id == 'DEF456'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_test(self, client, monkeypatch):
        response = {'user_info':{'user_id': 'foobar'}}

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        user = client.test()

        mock_make_request.assert_called_with('test.v1', {})

        assert isinstance(user, User)
        assert user.id == 'foobar'

    def test_release_reservation(self, client, monkeypatch):
        response = {'released_ok': True}

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        released = client.release_reservation('abc123')

        mock_make_request.assert_called_with('release.v1', {
            'transaction_uuid': 'abc123',
        }, method=POST)

        assert released is True

    def test_make_purchase_card_details(self, client, monkeypatch):
        response = {
            'transaction_status': 'purchased',
            'trolley_contents': {
                'transaction_uuid': 'DEF456'
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        customer = Customer('fred', 'flintstone', ['301 cobblestone way'], 'us')
        card_details = CardDetails(
            '4111 1111 1111 1111',
            expiry_year=17,
            expiry_month=3,
        )
        status, callout, meta = client.make_purchase(
            'abc123',
            customer,
            payment_method=card_details
        )

        expected_params = {
            'transaction_uuid': 'abc123',
            'first_name': 'fred',
            'last_name': 'flintstone',
            'address_line_one': '301 cobblestone way',
            'country_code': 'us',
            'card_number': '4111 1111 1111 1111',
            'expiry_date': '0317',
            'supplier_can_use_customer_data': False,
            'user_can_use_customer_data': False,
            'world_can_use_customer_data': False,
            'send_confirmation_email': True,
        }

        mock_make_request.assert_called_with(
            'purchase.v1',
            expected_params,
            method=POST
        )

        assert callout is None
        assert isinstance(status, Status)
        assert status.trolley.transaction_uuid == 'DEF456'
        assert status.status == 'purchased'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_make_purchase_redirection(self, client, monkeypatch):
        response = {
            "callout": {
                "bundle_source_code": "ext_test0",
            },
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        customer = Customer('fred', 'flintstone', ['301 cobblestone way'], 'us')

        redirection_details = RedirectionDetails(
            token='abc123',
            url='https://myticketingco.biz/confirmation/abc123',
            user_agent='Mozilla/5.0',
            accept='text/html,text/plain,application/json',
            remote_site='myticketingco.biz',
        )

        status, callout, meta = client.make_purchase(
            'abc123',
            customer,
            payment_method=redirection_details
        )

        expected_params = {
            'transaction_uuid': 'abc123',
            'first_name': 'fred',
            'last_name': 'flintstone',
            'address_line_one': '301 cobblestone way',
            'country_code': 'us',
            'return_token': 'abc123',
            'return_url': 'https://myticketingco.biz/confirmation/abc123',
            'client_http_user_agent': 'Mozilla/5.0',
            'client_http_accept': 'text/html,text/plain,application/json',
            'remote_site': 'myticketingco.biz',
            'supplier_can_use_customer_data': False,
            'user_can_use_customer_data': False,
            'world_can_use_customer_data': False,
            'send_confirmation_email': True,
        }

        mock_make_request.assert_called_with(
            'purchase.v1',
            expected_params,
            method=POST
        )

        assert status is None
        assert isinstance(callout, Callout)
        assert callout.code == 'ext_test0'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code is None

    def test_make_purchase_credit(self, client, monkeypatch):
        response = {
            'transaction_status': 'purchased',
            'trolley_contents': {
                'transaction_uuid': 'DEF456'
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        customer = Customer('fred', 'flintstone', ['301 cobblestone way'], 'us')

        status, callout, meta = client.make_purchase('abc123', customer)

        expected_params = {
            'transaction_uuid': 'abc123',
            'first_name': 'fred',
            'last_name': 'flintstone',
            'address_line_one': '301 cobblestone way',
            'country_code': 'us',
            'supplier_can_use_customer_data': False,
            'user_can_use_customer_data': False,
            'world_can_use_customer_data': False,
            'send_confirmation_email': True,
        }

        mock_make_request.assert_called_with(
            'purchase.v1',
            expected_params,
            method=POST
        )

        assert callout is None
        assert isinstance(status, Status)
        assert status.trolley.transaction_uuid == 'DEF456'
        assert status.status == 'purchased'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_make_purchase_opting_out_of_confirmation_email(self, client, monkeypatch):
        response = {
            'transaction_status': 'purchased',
            'trolley_contents': {
                'transaction_uuid': 'DEF456'
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        customer = Customer('fred', 'flintstone', ['301 cobblestone way'], 'us')

        status, callout, meta = client.make_purchase(
            'abc123',
            customer,
            send_confirmation_email=False
        )

        expected_params = {
            'transaction_uuid': 'abc123',
            'first_name': 'fred',
            'last_name': 'flintstone',
            'address_line_one': '301 cobblestone way',
            'country_code': 'us',
            'supplier_can_use_customer_data': False,
            'user_can_use_customer_data': False,
            'world_can_use_customer_data': False,
        }

        mock_make_request.assert_called_with(
            'purchase.v1',
            expected_params,
            method=POST
        )

        assert callout is None
        assert isinstance(status, Status)
        assert status.trolley.transaction_uuid == 'DEF456'
        assert status.status == 'purchased'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_get_purchase(self, client, monkeypatch):
        response = {
            'transaction_status': 'purchased',
            'trolley_contents': {
                'transaction_uuid': 'DEF456'
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        status, callout, meta = client.get_purchase('abc123')

        expected_params = {
            'transaction_uuid': 'abc123',
        }

        mock_make_request.assert_called_with(
            'purchase_page_archive.v1',
            expected_params,
            method=GET
        )

        assert callout is None
        assert isinstance(status, Status)
        assert status.trolley.transaction_uuid == 'DEF456'
        assert status.status == 'purchased'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_next_callout(self, client, monkeypatch):
        response = {
            'transaction_status': 'purchased',
            'trolley_contents': {
                'transaction_uuid': 'DEF456'
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        status, callout, meta = client.next_callout(
            'abc123',
            'def456',
            {'foo': 'bar'},
            lol='beans',
        )

        expected_params = {
            'foo': 'bar',
            'lol': 'beans',
        }

        mock_make_request.assert_called_with(
            'callback.v1/this.abc123/next.def456',
            expected_params,
            method=POST
        )

        assert callout is None
        assert isinstance(status, Status)
        assert status.trolley.transaction_uuid == 'DEF456'
        assert status.status == 'purchased'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_next_callout_with_additional_callout(self, client, monkeypatch):
        response = {
            "callout": {
                "bundle_source_code": "ext_test0",
            },
            'currency_code': 'gbp',
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        status, callout, meta = client.next_callout(
            'abc123',
            'def456',
            {'foo': 'bar'},
            lol='beans',
        )

        expected_params = {
            'foo': 'bar',
            'lol': 'beans',
        }

        mock_make_request.assert_called_with(
            'callback.v1/this.abc123/next.def456',
            expected_params,
            method=POST
        )

        assert status is None
        assert isinstance(callout, Callout)
        assert callout.code == 'ext_test0'

        assert 'gbp' in meta.currencies
        assert meta.default_currency_code == 'gbp'

    def test_auth_can_be_overridden_with_subclass(self, monkeypatch):
        """Test that we can override authentication behavior in subclasses

        Clients should be able to override the get_auth_params and make
        requests without basic authentication, if they can authenticate in
        another secure way.

        Since get_auth_params() has been deprecated, this should raise a
        DeprecationWarning, but still work (for legacy client support).
        """

        # state
        class MyClient(Client):
            def __init__(self, user, auth_key, **kwargs):
                super(MyClient, self).__init__(user, password=None, **kwargs)
                self.auth_key = auth_key

            def get_auth_params(self):
                return {
                    'user_id': self.user,
                    'auth_key': self.auth_key,
                }

        client = MyClient('gandalf', auth_key='speakfriendandenter',
                          use_decimal=True)

        params = {
            'foo': 'bar',
        }
        client.language='en-GB'

        # fakes
        fake_response = FakeResponse(status_code=200, json={"lol": "beans"})
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))

        # action
        with pytest.warns(DeprecationWarning) as warning_info:
            response = client.make_request('events.v1', params)

        # results
        assert response == {'lol': 'beans'}
        fake_get.assert_called_with(
            'https://api.ticketswitch.com/f13/events.v1/',
            auth=None,
            params={
                'foo': 'bar',
                'user_id': 'gandalf',
                'auth_key': 'speakfriendandenter',
            },
            headers={
                'Accept-Language': 'en-GB',
                'User-Agent': 'pyticketswitch {}'.format(pyticketswitch.__version__),
            },
            timeout=None
        )
        assert warning_info[0].message.args[0] == (
            'Function get_auth_params() is deprecated and should not be used')

    def test_extra_params_can_be_overriden_by_subclass(self, monkeypatch):
        """Test that we can override extra parameters in subclass

        Clients should be able to pass in extra parameters by overriding this
        method.
        """

        # state
        class MyClient(Client):
            def __init__(self, user, myfoo, **kwargs):
                super(MyClient, self).__init__(user, password=None, **kwargs)
                self.myfoo = myfoo

            def get_extra_params(self):
                params = super(MyClient, self).get_extra_params()
                params.update(myfoo=self.myfoo)
                return params

        client = MyClient('batman', 'batmanfoo',
                          sub_user='robin', use_decimal=True)
        params = {'fruit': 'apple'}

        # fakes
        fake_response = FakeResponse(status_code=200, json={'a': 'b'})
        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))

        # action
        response = client.make_request('events.v1', params)

        # results
        assert response == {'a': 'b'}
        fake_get.assert_called_with(
            'https://api.ticketswitch.com/f13/events.v1/',
            auth=None,
            params={
                'sub_id': 'robin',
                'myfoo': 'batmanfoo',
                'fruit': 'apple',
            },
            headers={
                'Accept-Language': 'en-GB',
                'User-Agent': 'pyticketswitch {}'.format(pyticketswitch.__version__),
            },
            timeout=None,
        )

    def test_get_auth_params_raises_deprecation_warning(self, client):
        """Tests that get_auth_params raises deprecation warning"""

        with pytest.warns(DeprecationWarning) as warning_list:
            params = client.get_auth_params()

        assert not params
        assert warning_list[0].message.args[0] == (
            'Call to deprecated function get_auth_params'
        )

    def test_make_request_using_decimal_parsing(self, client, monkeypatch):
        # fakes
        response_json = {'amount': 1.0}
        fake_response = requests.models.Response()
        fake_response._content = json.dumps(response_json).encode('utf-8')
        fake_response.status_code = 200

        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))

        # action
        result = client.make_request('test.v1', {})

        # results
        assert 'amount' in result
        assert type(result['amount']) == decimal.Decimal
        assert result['amount'] == decimal.Decimal('1.0')

    def test_make_request_using_float_parsing(self, monkeypatch):
        # state
        client = Client('bilbo', 'baggins')

        # fakes
        response_json = {'amount': 1.0}
        fake_response = requests.models.Response()
        fake_response._content = json.dumps(response_json).encode('utf-8')
        fake_response.status_code = 200

        fake_get = Mock(return_value=fake_response)
        session = Mock(spec=requests.Session)
        session.get = fake_get
        monkeypatch.setattr(client, 'get_session', Mock(return_value=session))

        # action
        result = client.make_request('test.v1', {})

        # results
        assert 'amount' in result
        assert type(result['amount']) == float
        assert result['amount'] == 1.0

    def test_make_purchase_with_agent_reference(self, client, monkeypatch):
        # state
        response = {
            "callout": {
                "bundle_source_code": "ext_test0",
            },
            'currency_details': {
                'gbp': {
                    'currency_code': 'gbp',
                }
            }
        }

        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        customer = Customer('fred', 'flintstone', ['301 cobblestone way'], 'us')

        redirection_details = RedirectionDetails(
            token='abc123',
            url='https://myticketingco.biz/confirmation/abc123',
            user_agent='Mozilla/5.0',
            accept='text/html,text/plain,application/json',
            remote_site='myticketingco.biz',
        )

        client.make_purchase(
            'abc123',
            customer,
            payment_method=redirection_details,
            agent_reference='myticketingco_ff01'
        )

        expected_params = {
            'transaction_uuid': 'abc123',
            'first_name': 'fred',
            'last_name': 'flintstone',
            'address_line_one': '301 cobblestone way',
            'country_code': 'us',
            'return_token': 'abc123',
            'return_url': 'https://myticketingco.biz/confirmation/abc123',
            'client_http_user_agent': 'Mozilla/5.0',
            'client_http_accept': 'text/html,text/plain,application/json',
            'remote_site': 'myticketingco.biz',
            'supplier_can_use_customer_data': False,
            'user_can_use_customer_data': False,
            'world_can_use_customer_data': False,
            'send_confirmation_email': True,
            'agent_reference': 'myticketingco_ff01',
        }

        mock_make_request.assert_called_with(
            'purchase.v1',
            expected_params,
            method=POST
        )

    def test_cancel_purchase(self, client, monkeypatch):
        # state
        with open("test_data/successful_cancellation.json", 'r') as file_handle:
            response = json.load(file_handle)
        mock_make_request = Mock(return_value=response)
        monkeypatch.setattr(client, 'make_request', mock_make_request)

        cancellation_result, meta = client.cancel_purchase('abc123', cancel_items_list=[1,2,3])

        assert cancellation_result.is_fully_cancelled()
        assert cancellation_result.cancelled_item_numbers == [1]
        assert 'gbp' in meta.currencies



from pyticketswitch.field import Field


class TestField:

    def test_from_api_data(self):

        data = {
            'custom_field_name': 'foo_bar',
            'custom_field_label': 'Foo Bar',
            'custom_field_data': 'this is a foo and a bar',
        }

        field = Field.from_api_data(data)
        assert field.name == 'foo_bar'
        assert field.label == 'Foo Bar'
        assert field.data == 'this is a foo and a bar'



from pyticketswitch.seat import Seat, SeatBlock


class TestSeatBlock:

    def test_from_api_data(self):

        data = ["D1.12"]

        seat_block = SeatBlock.from_api_data(data, separator='.', restricted_view_seats=[], seats_by_text_message={})

        assert seat_block.length == 1
        assert len(seat_block.seats) == 1
        assert seat_block.seats[0].id == 'D1.12'


class TestSeat:

    def test_from_api_data(self):

        data = {
            'col_id': '361',
            'full_id': 'GQ361',
            'is_restricted_view': True,
            'row_id': 'GQ',
            'seat_text_code': 'NEARTOILET',
            'separator': '-',
            'barcode': '0987654321',
            'seat_apple_wallet_urls': {
                'apple_wallet_gen_url': 'https://apple-wallet-api-dev.ticketswitch.io/generate_passes'
            },
            'seat_google_pay_urls': {
                'gpay_jwt_url': 'https://google-wallet-api-dev.ticketswitch.io/transactions/googlePayJWT',
                'gpay_save_url': 'https://google-wallet-api-dev.ticketswitch.io/transactions/saveToGooglePay'
            }
        }

        seat = Seat.from_api_data(data)

        assert seat.id == 'GQ361'
        assert seat.column == '361'
        assert seat.row == 'GQ'
        assert seat.separator == '-'
        assert seat.is_restricted is True
        assert seat.seat_text_code == 'NEARTOILET'
        assert seat.barcode == '0987654321'
        assert seat.seat_apple_wallet_urls.apple_wallet_gen_url == 'https://apple-wallet-api-dev.ticketswitch.io/generate_passes'
        assert seat.seat_google_pay_urls.gpay_jwt_url == 'https://google-wallet-api-dev.ticketswitch.io/transactions/googlePayJWT'
        assert seat.seat_google_pay_urls.gpay_save_url == 'https://google-wallet-api-dev.ticketswitch.io/transactions/saveToGooglePay'
        



from decimal import Decimal
from pyticketswitch.commission import Commission


class TestCommission:

    def test_from_api_data(self):

        data = {
            "amount_excluding_vat": 2.93,
            "amount_including_vat": 3.51,
            "commission_currency_code": "gbp",
        }

        commission = Commission.from_api_data(data)

        assert commission.including_vat == 3.51
        assert commission.excluding_vat == 2.93
        assert commission.currency_code == 'gbp'

    def test_from_api_data_with_decimal(self):

        data = {
            "amount_excluding_vat": Decimal('2.93'),
            "amount_including_vat": Decimal('3.51'),
            "commission_currency_code": "gbp",
        }

        commission = Commission.from_api_data(data)

        assert commission.including_vat == Decimal('3.51')
        assert commission.excluding_vat == Decimal('2.93')
        assert commission.currency_code == 'gbp'
